# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2020 Russell Johnson (russ4262) <russ4262@gmail.com>    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


from __future__ import print_function

__title__ = "Path Slot Operation"
__author__ = "russ4262 (Russell Johnson)"
__url__ = "http://www.freecadweb.org"
__doc__ = "Class and implementation of Slot operation."
__contributors__ = ""

import FreeCAD
from PySide import QtCore
import Path
import PathScripts.PathLog as PathLog
import PathScripts.PathUtils as PathUtils
import PathScripts.PathOp as PathOp
import math

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader
Part = LazyLoader('Part', globals(), 'Part')
Arcs = LazyLoader('draftgeoutils.arcs', globals(), 'draftgeoutils.arcs')
if FreeCAD.GuiUp:
    FreeCADGui = LazyLoader('FreeCADGui', globals(), 'FreeCADGui')


# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)


PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())
# PathLog.trackModule(PathLog.thisModule())


class ObjectSlot(PathOp.ObjectOp):
    '''Proxy object for Slot operation.'''

    def opFeatures(self, obj):
        '''opFeatures(obj) ... return all standard features'''
        return PathOp.FeatureTool | PathOp.FeatureDepths \
            | PathOp.FeatureHeights | PathOp.FeatureStepDown \
            | PathOp.FeatureCoolant | PathOp.FeatureBaseVertexes \
            | PathOp.FeatureBaseEdges | PathOp.FeatureBaseFaces

    def initOperation(self, obj):
        '''initOperation(obj) ... Initialize the operation by
        managing property creation and property editor status.'''
        self.propertiesReady = False

        self.initOpProperties(obj)  # Initialize operation-specific properties

        # For debugging
        if PathLog.getLevel(PathLog.thisModule()) != 4:
            obj.setEditorMode('ShowTempObjects', 2)  # hide

        if not hasattr(obj, 'DoNotSetDefaultValues'):
            self.opSetEditorModes(obj)

    def initOpProperties(self, obj, warn=False):
        '''initOpProperties(obj) ... create operation specific properties'''
        self.addNewProps = list()

        for (prtyp, nm, grp, tt) in self.opPropertyDefinitions():
            if not hasattr(obj, nm):
                obj.addProperty(prtyp, nm, grp, tt)
                self.addNewProps.append(nm)

        # Set enumeration lists for enumeration properties
        if len(self.addNewProps) > 0:
            ENUMS = self.opPropertyEnumerations()
            # ENUMS = self.getActiveEnumerations(obj)
            for n in ENUMS:
                if n in self.addNewProps:
                    setattr(obj, n, ENUMS[n])

            if warn:
                newPropMsg = translate('PathSlot', 'New property added to')
                newPropMsg += ' "{}": {}'.format(obj.Label, self.addNewProps) + '. '
                newPropMsg += translate('PathSlot', 'Check default value(s).')
                FreeCAD.Console.PrintWarning(newPropMsg + '\n')

        self.propertiesReady = True

    def opPropertyDefinitions(self):
        '''opPropertyDefinitions(obj) ... Store operation specific properties'''

        return [
            ("App::PropertyBool", "ShowTempObjects", "Debug",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Show the temporary path construction objects when module is in DEBUG mode.")),

            ("App::PropertyVectorDistance", "CustomPoint1", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Enter custom start point for slot path.")),
            ("App::PropertyVectorDistance", "CustomPoint2", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Enter custom end point for slot path.")),
            ("App::PropertyEnumeration", "CutPattern", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Set the geometric clearing pattern to use for the operation.")),
            ("App::PropertyDistance", "ExtendPathStart", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Positive extends the beginning of the path, negative shortens.")),
            ("App::PropertyDistance", "ExtendPathEnd", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Positive extends the end of the path, negative shortens.")),
            ("App::PropertyEnumeration", "LayerMode", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Complete the operation in a single pass at depth, or mulitiple passes to final depth.")),
            ("App::PropertyEnumeration", "PathOrientation", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Choose the path orientation with regard to the feature(s) selected.")),
            ("App::PropertyEnumeration", "Reference1", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Choose what point to use on the first selected feature.")),
            ("App::PropertyEnumeration", "Reference2", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Choose what point to use on the second selected feature.")),
            ("App::PropertyDistance", "ExtendRadius", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "For arcs/circlular edges, offset the radius for the path.")),
            ("App::PropertyBool", "ReverseDirection", "Slot",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Enable to reverse the cut direction of the slot path.")),

            ("App::PropertyVectorDistance", "StartPoint", "Start Point",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "The custom start point for the path of this operation")),
            ("App::PropertyBool", "UseStartPoint", "Start Point",
                QtCore.QT_TRANSLATE_NOOP("App::Property", "Make True, if specifying a Start Point"))
        ]

    def opPropertyEnumerations(self):
        # Enumeration lists for App::PropertyEnumeration properties
        return {
            'CutPattern': ['Line', 'ZigZag'],
            'LayerMode': ['Single-pass', 'Multi-pass'],
            'PathOrientation': ['Start to End', 'Perpendicular'],
            'Reference1': ['Center of Mass', 'Center of BoundBox',
                           'Lowest Point', 'Highest Point', 'Long Edge',
                           'Short Edge', 'Vertex'],
            'Reference2': ['Center of Mass', 'Center of BoundBox',
                           'Lowest Point', 'Highest Point', 'Vertex']
        }

    def opPropertyDefaults(self, obj, job):
        '''opPropertyDefaults(obj, job) ... returns a dictionary of default values
        for the operation's properties.'''
        defaults = {
            'CustomPoint1': FreeCAD.Vector(0.0, 0.0, 0.0),
            'ExtendPathStart': 0.0,
            'Reference1': 'Center of Mass',
            'CustomPoint2': FreeCAD.Vector(10.0, 10.0, 0.0),
            'ExtendPathEnd': 0.0,
            'Reference2': 'Center of Mass',
            'LayerMode': 'Multi-pass',
            'CutPattern': 'ZigZag',
            'PathOrientation': 'Start to End',
            'ExtendRadius': 0.0,
            'ReverseDirection': False,

            # For debugging
            'ShowTempObjects': False
        }

        return defaults

    def getActiveEnumerations(self, obj):
        """getActiveEnumerations(obj) ...
        Method returns dictionary of property enumerations based on
        active conditions in the operation."""
        ENUMS = self.opPropertyEnumerations()
        if hasattr(obj, 'Base'):
            if obj.Base:
                # (base, subsList) = obj.Base[0]
                subsList = obj.Base[0][1]
                subCnt = len(subsList)
                if subCnt == 1:
                    # Adjust available enumerations
                    ENUMS['Reference1'] = self._makeReference1Enumerations(subsList[0], True)
                elif subCnt == 2:
                    # Adjust available enumerations
                    ENUMS['Reference1'] = self._makeReference1Enumerations(subsList[0])
                    ENUMS['Reference2'] = self._makeReference2Enumerations(subsList[1])
        return ENUMS

    def updateEnumerations(self, obj):
        """updateEnumerations(obj) ...
        Method updates property enumerations based on active conditions
        in the operation.  Returns the updated enumerations dictionary.
        Existing property values must be stored, and then restored after
        the assignment of updated enumerations."""
        PathLog.debug('updateEnumerations()')
        # Save existing values
        pre_Ref1 = obj.Reference1
        pre_Ref2 = obj.Reference2

        # Update enumerations
        ENUMS = self.getActiveEnumerations(obj)
        obj.Reference1 = ENUMS['Reference1']
        obj.Reference2 = ENUMS['Reference2']

        # Restore pre-existing values if available with active enumerations.
        # If not, set to first element in active enumeration list.
        if pre_Ref1 in ENUMS['Reference1']:
            obj.Reference1 = pre_Ref1
        else:
            obj.Reference1 = ENUMS['Reference1'][0]
        if pre_Ref2 in ENUMS['Reference2']:
            obj.Reference2 = pre_Ref2
        else:
            obj.Reference2 = ENUMS['Reference2'][0]

        return ENUMS

    def opSetEditorModes(self, obj):
        # Used to hide inputs in properties list
        A = B = 2
        C = 0
        if hasattr(obj, 'Base'):
            if obj.Base:
                # (base, subsList) = obj.Base[0]
                subsList = obj.Base[0][1]
                subCnt = len(subsList)
                if subCnt == 1:
                    A = 0
                elif subCnt == 2:
                    A = B = 0
                    C = 2

        obj.setEditorMode('Reference1', A)
        obj.setEditorMode('Reference2', B)
        obj.setEditorMode('ExtendRadius', C)

    def onChanged(self, obj, prop):
        if hasattr(self, 'propertiesReady'):
            if self.propertiesReady:
                if prop in ['Base']:
                    self.updateEnumerations(obj)
                    self.opSetEditorModes(obj)

    def opOnDocumentRestored(self, obj):
        self.propertiesReady = False
        job = PathUtils.findParentJob(obj)

        self.initOpProperties(obj, warn=True)
        self.opApplyPropertyDefaults(obj, job, self.addNewProps)

        mode = 2 if PathLog.getLevel(PathLog.thisModule()) != 4 else 0
        obj.setEditorMode('ShowTempObjects', mode)

        # Repopulate enumerations in case of changes
        ENUMS = self.updateEnumerations(obj)
        for n in ENUMS:
            restore = False
            if hasattr(obj, n):
                val = obj.getPropertyByName(n)
                restore = True
            setattr(obj, n, ENUMS[n])  # set the enumerations list
            if restore:
                setattr(obj, n, val)  # restore the value

        self.opSetEditorModes(obj)

    def opApplyPropertyDefaults(self, obj, job, propList):
        # Set standard property defaults
        PROP_DFLTS = self.opPropertyDefaults(obj, job)
        for n in PROP_DFLTS:
            if n in propList:
                prop = getattr(obj, n)
                val = PROP_DFLTS[n]
                setVal = False
                if hasattr(prop, 'Value'):
                    if isinstance(val, int) or isinstance(val, float):
                        setVal = True
                if setVal:
                    # propVal = getattr(prop, 'Value')
                    setattr(prop, 'Value', val)
                else:
                    setattr(obj, n, val)

    def opSetDefaultValues(self, obj, job):
        '''opSetDefaultValues(obj, job) ... initialize defaults'''
        job = PathUtils.findParentJob(obj)

        self.opApplyPropertyDefaults(obj, job, self.addNewProps)

        # need to overwrite the default depth calculations for facing
        d = None
        if job:
            if job.Stock:
                d = PathUtils.guessDepths(job.Stock.Shape, None)
                PathLog.debug("job.Stock exists")
            else:
                PathLog.debug("job.Stock NOT exist")
        else:
            PathLog.debug("job NOT exist")

        if d is not None:
            obj.OpFinalDepth.Value = d.final_depth
            obj.OpStartDepth.Value = d.start_depth
        else:
            obj.OpFinalDepth.Value = -10
            obj.OpStartDepth.Value = 10

        PathLog.debug('Default OpFinalDepth: {}'.format(obj.OpFinalDepth.Value))
        PathLog.debug('Defualt OpStartDepth: {}'.format(obj.OpStartDepth.Value))

    def opApplyPropertyLimits(self, obj):
        '''opApplyPropertyLimits(obj) ... Apply necessary limits to user input property values before performing main operation.'''
        pass

    def opUpdateDepths(self, obj):
        if hasattr(obj, 'Base') and obj.Base:
            base, sublist = obj.Base[0]
            fbb = base.Shape.getElement(sublist[0]).BoundBox
            zmin = fbb.ZMax
            for base, sublist in obj.Base:
                for sub in sublist:
                    try:
                        fbb = base.Shape.getElement(sub).BoundBox
                        zmin = min(zmin, fbb.ZMin)
                    except Part.OCCError as e:
                        PathLog.error(e)
            obj.OpFinalDepth = zmin

    def opExecute(self, obj):
        '''opExecute(obj) ... process surface operation'''
        PathLog.track()

        self.base = None
        self.shape1 = None
        self.shape2 = None
        self.shapeType1 = None
        self.shapeType2 = None
        self.shapeLength1 = None
        self.shapeLength2 = None
        self.dYdX1 = None
        self.dYdX2 = None
        self.bottomEdges = None
        self.stockZMin = None
        self.isArc = 0
        self.arcCenter = None
        self.arcMidPnt = None
        self.arcRadius = 0.0
        self.newRadius = 0.0
        self.isDebug = False if PathLog.getLevel(PathLog.thisModule()) != 4 else True
        self.showDebugObjects = False
        self.stockZMin = self.job.Stock.Shape.BoundBox.ZMin
        CMDS = list()

        try:
            dotIdx = __name__.index('.') + 1
        except Exception:
            dotIdx = 0
        self.module = __name__[dotIdx:]

        # Setup debugging group for temp objects, when in DEBUG mode
        if self.isDebug:
            self.showDebugObjects = obj.ShowTempObjects
        if self.showDebugObjects:
            FCAD = FreeCAD.ActiveDocument
            for grpNm in ['tmpDebugGrp', 'tmpDebugGrp001']:
                if hasattr(FCAD, grpNm):
                    for go in FCAD.getObject(grpNm).Group:
                        FCAD.removeObject(go.Name)
                    FCAD.removeObject(grpNm)
            self.tmpGrp = FCAD.addObject('App::DocumentObjectGroup', 'tmpDebugGrp')

        # Begin GCode for operation with basic information
        # ... and move cutter to clearance height and startpoint
        tool = obj.ToolController.Tool
        toolType = tool.ToolType if hasattr(tool, 'ToolType') else tool.ShapeName
        output = ''
        if obj.Comment != '':
            self.commandlist.append(Path.Command('N ({})'.format(obj.Comment), {}))
        self.commandlist.append(Path.Command('N ({})'.format(obj.Label), {}))
        self.commandlist.append(Path.Command('N (Tool type: {})'.format(toolType), {}))
        self.commandlist.append(Path.Command('N (Compensated Tool Path. Diameter: {})'.format(tool.Diameter), {}))
        self.commandlist.append(Path.Command('N ({})'.format(output), {}))
        self.commandlist.append(Path.Command('G0', {'Z': obj.ClearanceHeight.Value, 'F': self.vertRapid}))
        if obj.UseStartPoint is True:
            self.commandlist.append(Path.Command('G0', {'X': obj.StartPoint.x, 'Y': obj.StartPoint.y, 'F': self.horizRapid}))

        # Impose property limits
        self.opApplyPropertyLimits(obj)

        # Calculate default depthparams for operation
        self.depthParams = PathUtils.depth_params(obj.ClearanceHeight.Value, obj.SafeHeight.Value, obj.StartDepth.Value, obj.StepDown.Value, 0.0, obj.FinalDepth.Value)

        # ######  MAIN COMMANDS FOR OPERATION ######

        cmds = self._makeOperation(obj)
        if cmds:
            CMDS.extend(cmds)

        # Save gcode produced
        CMDS.append(Path.Command('G0', {'Z': obj.ClearanceHeight.Value, 'F': self.vertRapid}))
        self.commandlist.extend(CMDS)

        # ######  CLOSING COMMANDS FOR OPERATION ######

        # Hide the temporary objects
        if self.showDebugObjects:
            if FreeCAD.GuiUp:
                FreeCADGui.ActiveDocument.getObject(self.tmpGrp.Name).Visibility = False
            self.tmpGrp.purgeTouched()

        return True

    # Control methods for operation
    def _makeOperation(self, obj):
        """This method controls the overall slot creation process."""
        pnts = False
        featureCount = 0

        if not hasattr(obj, 'Base'):
            msg = translate('PathSlot',
                    'No Base Geometry object in the operation.')
            FreeCAD.Console.PrintError(msg + '\n')
            return False

        if not obj.Base:
            # Use custom inputs here
            p1 = obj.CustomPoint1
            p2 = obj.CustomPoint2
            if p1.z == p2.z:
                pnts = (p1, p2)
            else:
                msg = translate('PathSlot',
                        'Custom points not at same Z height.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False

        baseGeom = obj.Base[0]
        base, subsList = baseGeom
        self.base = base

        featureCount = len(subsList)
        if  featureCount == 1:
            PathLog.debug('Reference 1: {}'.format(obj.Reference1))
            sub1 = subsList[0]
            shape_1 = getattr(base.Shape, sub1)
            self.shape1 = shape_1
            pnts = self._processSingle(obj, shape_1, sub1)
        else:
            PathLog.debug('Reference 1: {}'.format(obj.Reference1))
            PathLog.debug('Reference 2: {}'.format(obj.Reference2))
            sub1 = subsList[0]
            sub2 = subsList[1]
            shape_1 = getattr(base.Shape, sub1)
            shape_2 = getattr(base.Shape, sub2)
            self.shape1 = shape_1
            self.shape2 = shape_2
            pnts = self._processDouble(obj, shape_1, sub1, shape_2, sub2)

        if not pnts:
            return False

        if self.isArc:
            cmds = self._finishArc(obj, pnts, featureCount)
        else:
            cmds = self._finishLine(obj, pnts, featureCount)

        if cmds:
            return cmds

        return False

    def _finishArc(self, obj, pnts, featureCnt):
        """This method finishes an Arc Slot operation.
        It returns the gcode for the slot operation."""
        PathLog.debug('arc center: {}'.format(self.arcCenter))
        self._addDebugObject(Part.makeLine(self.arcCenter, self.arcMidPnt), 'CentToMidPnt')

        # PathLog.debug('Pre-offset points are:\np1 = {}\np2 = {}'.format(p1, p2))
        if obj.ExtendRadius.Value != 0:
            # verify offset does not force radius < 0
            newRadius = self.arcRadius + obj.ExtendRadius.Value
            PathLog.debug('arc radius: {};  offset radius: {}'.format(self.arcRadius, newRadius))
            if newRadius <= 0:
                msg = translate('PathSlot',
                        'Current Extend Radius value produces negative arc radius.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False
            else:
                (p1, p2) = pnts
                pnts = self._makeOffsetArc(p1, p2, self.arcCenter, newRadius)
                self.newRadius = newRadius
        else:
            PathLog.debug('arc radius: {}'.format(self.arcRadius))
            self.newRadius = self.arcRadius

        # Apply path extension for arcs
        # PathLog.debug('Pre-extension points are:\np1 = {}\np2 = {}'.format(p1, p2))
        if self.isArc == 1:
            # Complete circle
            if (obj.ExtendPathStart.Value != 0 or
                obj.ExtendPathEnd.Value != 0):
                msg = translate('PathSlot',
                        'No path extensions available for full circles.')
                FreeCAD.Console.PrintWarning(msg + '\n')
        else:
            # Arc segment
            # Apply extensions to slot path
            (p1, p2) = pnts
            begExt = obj.ExtendPathStart.Value
            endExt = obj.ExtendPathEnd.Value
            # invert endExt, begExt args to apply extensions to correct ends
            # XY geom is postitive CCW; Gcode postitive CW
            pnts = self._extendArcSlot(p1, p2, self.arcCenter, endExt, begExt)

        if not pnts:
            return False

        (p1, p2) = pnts
        # PathLog.error('Post-offset points are:\np1 = {}\np2 = {}'.format(p1, p2))
        if self.isDebug:
            PathLog.debug('Path Points are:\np1 = {}\np2 = {}'.format(p1, p2))
            if p1.sub(p2).Length != 0:
                self._addDebugObject(Part.makeLine(p1, p2), 'Path')

        if featureCnt:
            obj.CustomPoint1 = p1
            obj.CustomPoint2 = p2

        if self._arcCollisionCheck(obj, p1, p2, self.arcCenter, self.newRadius):
            msg = obj.Label + ' '
            msg += translate('PathSlot',
                    'operation collides with model.')
            FreeCAD.Console.PrintError(msg + '\n')

        # PathLog.warning('Unable to create G-code.  _makeArcGCode() is incomplete.')
        cmds = self._makeArcGCode(obj, p1, p2)
        return cmds

    def _makeArcGCode(self, obj, p1, p2):
        """This method is the last step in the overall arc slot creation process.
        It accepts the operation object and two end points for the path.
        It returns the gcode for the slot operation."""
        CMDS = list()
        PATHS = [(p2, p1, 'G2'), (p1, p2, 'G3')]
        if obj.ReverseDirection :
            path_index = 1 
        else : 
            path_index = 0

        def arcPass(POINTS, depth):
            cmds = list()
            (st_pt, end_pt, arcCmd) = POINTS
            # cmds.append(Path.Command('N (Tool type: {})'.format(toolType), {}))
            cmds.append(Path.Command('G0', {'X': st_pt.x, 'Y': st_pt.y, 'F': self.horizRapid}))
            cmds.append(Path.Command('G1', {'Z': depth, 'F': self.vertFeed}))
            vtc = self.arcCenter.sub(st_pt)  # vector to center
            cmds.append(
                Path.Command(arcCmd,
                    {'X': end_pt.x, 'Y': end_pt.y, 'I': vtc.x,
                        'J': vtc.y, 'F': self.horizFeed
                     }))
            return cmds

        if obj.LayerMode == 'Single-pass':
            CMDS.extend(arcPass(PATHS[path_index], obj.FinalDepth.Value))
        else:
            if obj.CutPattern == 'Line':
                for depth in self.depthParams:
                    CMDS.extend(arcPass(PATHS[path_index], depth))
                    CMDS.append(Path.Command('G0', {'Z': obj.SafeHeight.Value, 'F': self.vertRapid}))
            elif obj.CutPattern == 'ZigZag':
                i = 0
                for depth in self.depthParams:
                    if i % 2.0 == 0:  # even
                        CMDS.extend(arcPass(PATHS[path_index], depth))
                    else:  # odd
                        CMDS.extend(arcPass(PATHS[not path_index], depth))
                    i += 1
        # Raise to SafeHeight when finished
        CMDS.append(Path.Command('G0', {'Z': obj.SafeHeight.Value, 'F': self.vertRapid}))

        if self.isDebug:
            PathLog.debug('G-code arc command is: {}'.format(PATHS[path_index][2]))

        return CMDS

    def _finishLine(self, obj, pnts, featureCnt):
        """This method finishes a Line Slot operation.
        It returns the gcode for the line slot operation."""
        # Apply perpendicular rotation if requested
        perpZero = True
        if obj.PathOrientation == 'Perpendicular':
            if featureCnt == 2:
                if self.shapeType1 == 'Face' and self.shapeType2 == 'Face':
                    if self.bottomEdges:
                        self.bottomEdges.sort(key=lambda edg: edg.Length, reverse=True)
                        BE = self.bottomEdges[0]
                        pnts = self._processSingleVertFace(obj, BE)
                        perpZero = False
                elif self.shapeType1 == 'Edge' and self.shapeType2 == 'Edge':
                    PathLog.debug('_finishLine() Perp, featureCnt == 2')
            if perpZero:
                (p1, p2) = pnts
                pnts = self._makePerpendicular(p1, p2, 10.0)  # 10.0 offset below
        else:
            # Modify path points if user selected two parallel edges
            if (featureCnt == 2 and self.shapeType1 == 'Edge' and 
                    self.shapeType2 == 'Edge' and self._isParallel(self.dYdX1, self.dYdX2)):
                (p1, p2) = pnts
                edg1_len = self.shape1.Length
                edg2_len = self.shape2.Length
                set_length = max(edg1_len, edg2_len)
                pnts = self._makePerpendicular(p1, p2, 10.0 + set_length)  # 10.0 offset below
                if edg1_len != edg2_len:
                    msg = obj.Label + ' '
                    msg += translate('PathSlot',
                            'Verify slot path start and end points.')
                    FreeCAD.Console.PrintWarning(msg + '\n')
            else:
                perpZero = False

        # Reverse direction of path if requested
        if obj.ReverseDirection:
            (p2, p1) = pnts
        else:
            (p1, p2) = pnts

        # Apply extensions to slot path
        begExt = obj.ExtendPathStart.Value
        endExt = obj.ExtendPathEnd.Value
        if perpZero:
            # Offsets for 10.0 value above in _makePerpendicular()
            begExt -= 5.0
            endExt -= 5.0
        pnts = self._extendLineSlot(p1, p2, begExt, endExt)

        if not pnts:
            return False

        (p1, p2) = pnts
        if self.isDebug:
            PathLog.debug('Path Points are:\np1 = {}\np2 = {}'.format(p1, p2))
            if p1.sub(p2).Length != 0:
                self._addDebugObject(Part.makeLine(p1, p2), 'Path')

        if featureCnt:
            obj.CustomPoint1 = p1
            obj.CustomPoint2 = p2

        if self._lineCollisionCheck(obj, p1, p2):
            msg = obj.Label + ' '
            msg += translate('PathSlot',
                    'operation collides with model.')
            FreeCAD.Console.PrintWarning(msg + '\n')

        cmds = self._makeLineGCode(obj, p1, p2)
        return cmds

    def _makeLineGCode(self, obj, p1, p2):
        """This method is the last in the overall line slot creation process.
        It accepts the operation object and two end points for the path.
        It returns the gcode for the slot operation."""
        CMDS = list()

        def linePass(p1, p2, depth):
            cmds = list()
            # cmds.append(Path.Command('N (Tool type: {})'.format(toolType), {}))
            cmds.append(Path.Command('G0', {'X': p1.x, 'Y': p1.y, 'F': self.horizRapid}))
            cmds.append(Path.Command('G1', {'Z': depth, 'F': self.vertFeed}))
            cmds.append(Path.Command('G1', {'X': p2.x, 'Y': p2.y, 'F': self.horizFeed}))
            return cmds

        # CMDS.append(Path.Command('N (Tool type: {})'.format(toolType), {}))
        if obj.LayerMode == 'Single-pass':
            CMDS.extend(linePass(p1, p2, obj.FinalDepth.Value))
            CMDS.append(Path.Command('G0', {'Z': obj.SafeHeight.Value, 'F': self.vertRapid}))
        else:
            if obj.CutPattern == 'Line':
                for dep in self.depthParams:
                    CMDS.extend(linePass(p1, p2, dep))
                    CMDS.append(Path.Command('G0', {'Z': obj.SafeHeight.Value, 'F': self.vertRapid}))
            elif obj.CutPattern == 'ZigZag':
                CMDS.append(Path.Command('G0', {'X': p1.x, 'Y': p1.y, 'F': self.horizRapid}))
                i = 0
                for dep in self.depthParams:
                    if i % 2.0 == 0:  # even
                        CMDS.append(Path.Command('G1', {'Z': dep, 'F': self.vertFeed}))
                        CMDS.append(Path.Command('G1', {'X': p2.x, 'Y': p2.y, 'F': self.horizFeed}))
                    else:  # odd
                        CMDS.append(Path.Command('G1', {'Z': dep, 'F': self.vertFeed}))
                        CMDS.append(Path.Command('G1', {'X': p1.x, 'Y': p1.y, 'F': self.horizFeed}))
                    i += 1
            CMDS.append(Path.Command('G0', {'Z': obj.SafeHeight.Value, 'F': self.vertRapid}))

        return CMDS

    # Methods for processing single geometry
    def _processSingle(self, obj, shape_1, sub1):
        """This is the control method for slots based on a
        single Base Geometry feature."""
        done = False
        cat1 = sub1[:4]

        if cat1 == 'Face':
            pnts = False
            norm = shape_1.normalAt(0.0, 0.0)
            PathLog.debug('{}.normalAt(): {}'.format(sub1, norm))
            if norm.z == 1 or norm.z == -1:
                pnts = self._processSingleHorizFace(obj, shape_1)
            elif norm.z == 0:
                faceType = self._getVertFaceType(shape_1)
                if faceType:
                    (geo, shp) = faceType
                    if geo == 'Face':
                        pnts = self._processSingleComplexFace(obj, shp)
                    if geo == 'Wire':
                        pnts = self._processSingleVertFace(obj, shp)
                    if geo == 'Edge':
                        pnts = self._processSingleVertFace(obj, shp)
            else:
                msg = translate('PathSlot',
                    'The selected face is not oriented horizontally or vertically.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False

            if pnts:
                (p1, p2) = pnts
                done = True

        elif cat1 == 'Edge':
            PathLog.debug('Single edge')
            pnts = self._processSingleEdge(obj, shape_1)
            if pnts:
                (p1, p2) = pnts
                done = True

        elif cat1 == 'Vert':
            msg = translate('PathSlot',
                'Only a vertex selected. Add another feature to the Base Geometry.')
            FreeCAD.Console.PrintError(msg + '\n')

        if done:
            return (p1, p2)

        return False

    def _processSingleHorizFace(self, obj, shape):
        """Determine slot path endpoints from a single horizontally oriented face."""
        PathLog.debug('_processSingleHorizFace()')
        lineTypes = ['Part::GeomLine']

        def getRadians(self, E):
            vect = self._dXdYdZ(E)
            norm = self._normalizeVector(vect)
            rads = self._getVectorAngle(norm)
            deg = math.degrees(rads)
            if deg >= 180.0:
                deg -= 180.0
            return deg

        # Reject triangular faces
        if len(shape.Edges) < 4:
            msg = translate('PathSlot',
                'A single selected face must have four edges minimum.')
            FreeCAD.Console.PrintError(msg + '\n')
            return False

        # Create tuples as (edge index, length, angle)
        eTups = list()
        for i in range(0, 4):
            eTups.append((i,
                             shape.Edges[i].Length,
                             getRadians(self, shape.Edges[i]))
                            )

        # Sort tuples by edge angle
        eTups.sort(key=lambda tup: tup[2])

        # Identify parallel edges
        parallel_edge_pairs = list()
        parallel_edge_flags = list()
        flag = 1
        eCnt = len(shape.Edges)
        lstE = eCnt - 1
        for i in range(0, eCnt):  # populate empty parallel edge flag list
            parallel_edge_flags.append(0)        
        for i in range(0, eCnt):  # Cycle through edges to identify parallel pairs
            if i < lstE:
                ni = i + 1
                A = eTups[i]
                B = eTups[ni]
                if abs(A[2] - B[2]) < 0.00000001:  # test slopes(yaw angles)
                    debug = False
                    eA = shape.Edges[A[0]]
                    eB = shape.Edges[B[0]]
                    if eA.Curve.TypeId not in lineTypes:
                        debug = eA.Curve.TypeId
                    if not debug:
                        if eB.Curve.TypeId not in lineTypes:
                            debug = eB.Curve.TypeId
                        else:
                            parallel_edge_pairs.append((eA, eB))
                            # set parallel flags for this pair of edges
                            parallel_edge_flags[A[0]] = flag
                            parallel_edge_flags[B[0]] = flag
                            flag += 1
                    if debug:
                        msg = 'Erroneous Curve.TypeId: {}'.format(debug)
                        PathLog.debug(msg)

        pairCnt = len(parallel_edge_pairs)
        if pairCnt > 1:
            parallel_edge_pairs.sort(key=lambda tup: tup[0].Length, reverse=True)

        if self.isDebug:
            PathLog.debug(' -pairCnt: {}'.format(pairCnt))
            for (a, b) in parallel_edge_pairs:
                PathLog.debug(' -pair: {}, {}'.format(round(a.Length, 4), round(b.Length,4)))
            PathLog.debug(' -parallel_edge_flags: {}'.format(parallel_edge_flags))

        if pairCnt == 0:
            msg = translate('PathSlot',
                'No parallel edges identified.')
            FreeCAD.Console.PrintError(msg + '\n')
            return False
        elif pairCnt == 1:
            # One pair of parallel edges identified
            if eCnt == 4:
                flag_set = list()
                for i in range(0, 4):
                    e = parallel_edge_flags[i]
                    if e == 0:
                        flag_set.append(shape.Edges[i])
                if len(flag_set) == 2:
                    same = (flag_set[0], flag_set[1])
                else:
                    same = parallel_edge_pairs[0]
            else:
                same = parallel_edge_pairs[0]
        else:
            if obj.Reference1 == 'Long Edge':
                same = parallel_edge_pairs[1]
            elif obj.Reference1 == 'Short Edge':
                same = parallel_edge_pairs[0]
            else:
                msg = 'Reference1 '
                msg += translate('PathSlot',
                    'value error.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False

        (p1, p2) = self._getOppMidPoints(same)
        return (p1, p2)

    def _processSingleComplexFace(self, obj, shape):
        """Determine slot path endpoints from a single complex face."""
        PathLog.debug('_processSingleComplexFace()')
        PNTS = list()

        def zVal(V):
            return V.z

        for E in shape.Wires[0].Edges:
            p = self._findLowestEdgePoint(E)
            PNTS.append(p)
        PNTS.sort(key=zVal)
        return (PNTS[0], PNTS[1])

    def _processSingleVertFace(self, obj, shape):
        """Determine slot path endpoints from a single vertically oriented face
        with no single bottom edge."""
        PathLog.debug('_processSingleVertFace()')
        eCnt = len(shape.Edges)
        V0 = shape.Edges[0].Vertexes[0]
        V1 = shape.Edges[eCnt - 1].Vertexes[1]
        v0 = FreeCAD.Vector(V0.X, V0.Y, V0.Z)
        v1 = FreeCAD.Vector(V1.X, V1.Y, V1.Z)

        dX = V1.X - V0.X
        dY = V1.Y - V0.Y
        dZ = V1.Z - V0.Z
        temp = FreeCAD.Vector(dX, dY, dZ)
        slope = self._normalizeVector(temp)
        perpVect = FreeCAD.Vector(-1 * slope.y, slope.x, slope.z)
        perpVect.multiply(self.tool.Diameter / 2.0)

        # Create offset endpoints for raw slot path
        a1 = v0.add(perpVect)
        a2 = v1.add(perpVect)
        b1 = v0.sub(perpVect)
        b2 = v1.sub(perpVect)
        (p1, p2) = self._getCutSidePoints(obj, v0, v1, a1, a2, b1, b2)

        msg = obj.Label + ' '
        msg += translate('PathSlot',
                'Verify slot path start and end points.')
        FreeCAD.Console.PrintWarning(msg + '\n')

        return (p1, p2)

    def _processSingleEdge(self, obj, edge):
        """Determine slot path endpoints from a single horizontally oriented edge."""
        PathLog.debug('_processSingleEdge()')
        tolrnc = 0.0000001
        lineTypes = ['Part::GeomLine']
        curveTypes = ['Part::GeomCircle']

        def oversizedTool(holeDiam):
            # Test if tool larger than opening
            if self.tool.Diameter > holeDiam:
                msg = translate('PathSlot',
                        'Current tool larger than arc diameter.')
                FreeCAD.Console.PrintError(msg + '\n')
                return True
            return False

        def isHorizontal(z1, z2, z3):
            # Check that all Z values are equal (isRoughly same)
            if (abs(z1 - z2) > tolrnc or
                abs(z1 - z3) > tolrnc ):
#                abs(z2 - z3) > tolrnc):  3rd test redundant.
                return False
            return True

        def circumCircleFrom3Points(P1, P2, P3):
            # Source code for this function copied from (with modifications):
            # https://wiki.freecadweb.org/Macro_Draft_Circle_3_Points_3D
            vP2P1 = (P2 - P1)
            vP3P2 = (P3 - P2)
            vP1P3 = (P1 - P3)

            L = vP2P1.cross(vP3P2).Length
            # Circle radius (not used)
            # r = vP1P2.Length * vP2P3.Length * vP3P1.Length / 2 / l
            if round(L, 8) == 0.0:
                PathLog.error("The three points are colinear, arc is a straight.")
                return False

            # Sphere center.
            twolsqr= 2*L*L
            a = -vP3P2.dot(vP3P2) * vP2P1.dot(vP1P3) / twolsqr
            b = -vP1P3.dot(vP1P3) * vP3P2.dot(vP2P1) / twolsqr
            c = -vP2P1.dot(vP2P1) * vP1P3.dot(vP3P2) / twolsqr
            return P1*a + P2*b + P3*c

        V1 = edge.Vertexes[0]
        p1 = FreeCAD.Vector(V1.X, V1.Y, 0.0)
        if len(edge.Vertexes) == 1:   # circle has one virtex
            p2 = FreeCAD.Vector(p1)
        else :  
            V2 = edge.Vertexes[1]
            p2 = FreeCAD.Vector(V2.X, V2.Y, 0.0)
         
        # Process edge based on curve type
        if edge.Curve.TypeId in lineTypes:
            return (p1, p2)
            
        elif edge.Curve.TypeId in curveTypes:
            if len(edge.Vertexes) == 1:
                # Circle edge
                PathLog.debug('Arc with single vertex.')
                if oversizedTool(edge.BoundBox.XLength):
                    return False

                self.isArc = 1
                tp1 = edge.valueAt(edge.getParameterByLength(edge.Length * 0.33))
                tp2 = edge.valueAt(edge.getParameterByLength(edge.Length * 0.66))
                if not isHorizontal(V1.Z, tp1.z, tp2.z):
                    return False

                center = edge.BoundBox.Center
                self.arcCenter = FreeCAD.Vector(center.x, center.y, 0.0)
                midPnt = edge.valueAt(edge.getParameterByLength(edge.Length / 2.0))
                self.arcMidPnt = FreeCAD.Vector(midPnt.x, midPnt.y, 0.0)
                self.arcRadius = edge.BoundBox.XLength / 2.0
            else:
                # Arc edge
                PathLog.debug('Arc with multiple vertices.')
                self.isArc = 2
                midPnt = edge.valueAt(edge.getParameterByLength(edge.Length / 2.0))
                if not isHorizontal(V1.Z, V2.Z, midPnt.z):
                    return False
                    
                midPnt.z = 0.0
                circleCenter = circumCircleFrom3Points(p1, p2, midPnt)
                if not circleCenter:
                    return False
                self.arcMidPnt = midPnt
                self.arcCenter = circleCenter
                self.arcRadius = p1.sub(circleCenter).Length

                if oversizedTool(self.arcRadius * 2.0):
                    return False

            return (p1, p2)
        else :
            msg = translate('PathSlot','Failed, slot from edge only accepts lines, arcs and circles.')
            FreeCAD.Console.PrintError(msg + '\n')
           
            return False  # not line , not circle

    # Methods for processing double geometry
    def _processDouble(self, obj, shape_1, sub1, shape_2, sub2):
        """This is the control method for slots based on a
        two Base Geometry features."""
        PathLog.debug('_processDouble()')

        p1 = None
        p2 = None
        dYdX1 = None
        dYdX2 = None
        self.bottomEdges = list()

        feature1 = self._processFeature(obj, shape_1, sub1, 1)
        if not feature1:
            msg = translate('PathSlot',
                'Failed to determine point 1 from')
            FreeCAD.Console.PrintError(msg + ' {}.\n'.format(sub1))
            return False
        (p1, dYdX1, shpType) = feature1
        self.shapeType1 = shpType
        if dYdX1:
            self.dYdX1 = dYdX1

        feature2 = self._processFeature(obj, shape_2, sub2, 2)
        if not feature2:
            msg = translate('PathSlot',
                'Failed to determine point 2 from')
            FreeCAD.Console.PrintError(msg + ' {}.\n'.format(sub2))
            return False
        (p2, dYdX2, shpType) = feature2
        self.shapeType2 = shpType
        if dYdX2:
            self.dYdX2 = dYdX2

        # Parallel check for twin face, and face-edge cases
        if dYdX1 and dYdX2:
            PathLog.debug('dYdX1, dYdX2: {}, {}'.format(dYdX1, dYdX2))
            if not self._isParallel(dYdX1, dYdX2):
                if self.shapeType1 != 'Edge' or self.shapeType2 != 'Edge':
                    msg = translate('PathSlot',
                        'Selected geometry not parallel.')
                    FreeCAD.Console.PrintError(msg + '\n')
                    return False

        if p2:
            return (p1, p2)

        return False

    # Support methods
    def _dXdYdZ(self, E):
        v1 = E.Vertexes[0]
        v2 = E.Vertexes[1]
        dX = v2.X - v1.X
        dY = v2.Y - v1.Y
        dZ = v2.Z - v1.Z
        return FreeCAD.Vector(dX, dY, dZ)

    def _normalizeVector(self, v):
        """_normalizeVector(v)...
        Returns a copy of the vector received with values rounded to 10 decimal places."""
        posTol = 0.0000000001  # abitrary, use job Geometry Tolerance ???
        negTol = -1 * posTol
        V = FreeCAD.Vector(v.x, v.y, v.z)
        V.normalize()
        x = V.x
        y = V.y
        z = V.z

        if V.x != 0 and abs(V.x) < posTol:
            x = 0.0
        if V.x != 1 and 1.0 - V.x < posTol:
            x = 1.0
        if V.x != -1 and -1.0 - V.x > negTol:
            x = -1.0

        if V.y != 0 and abs(V.y) < posTol:
            y = 0.0
        if V.y != 1 and 1.0 - V.y < posTol:
            y = 1.0
        if V.y != -1 and -1.0 - V.y > negTol:
            y = -1.0

        if V.z != 0 and abs(V.z) < posTol:
            z = 0.0
        if V.z != 1 and 1.0 - V.z < posTol:
            z = 1.0
        if V.z != -1 and -1.0 - V.z > negTol:
            z = -1.0

        return FreeCAD.Vector(x, y, z)

    def _getLowestPoint(self, shape_1):
        """_getLowestPoint(shape)... Returns lowest vertex of shape as vector."""
        # find lowest vertex
        vMin = shape_1.Vertexes[0]
        zmin = vMin.Z
        same = [vMin]
        for V in shape_1.Vertexes:
            if V.Z < zmin:
                zmin = V.Z
                # vMin = V
            elif V.Z == zmin:
                same.append(V)
        if len(same) > 1:
            X = [E.X for E in same]
            Y = [E.Y for E in same]
            avgX = sum(X) / len(X)
            avgY = sum(Y) / len(Y)
            return FreeCAD.Vector(avgX, avgY, zmin)
        else:
            return FreeCAD.Vector(V.X, V.Y, V.Z)

    def _getHighestPoint(self, shape_1):
        """_getHighestPoint(shape)... Returns highest vertex of shape as vector."""
        # find highest vertex
        vMax = shape_1.Vertexes[0]
        zmax = vMax.Z
        same = [vMax]
        for V in shape_1.Vertexes:
            if V.Z > zmax:
                zmax = V.Z
                # vMax = V
            elif V.Z == zmax:
                same.append(V)
        if len(same) > 1:
            X = [E.X for E in same]
            Y = [E.Y for E in same]
            avgX = sum(X) / len(X)
            avgY = sum(Y) / len(Y)
            return FreeCAD.Vector(avgX, avgY, zmax)
        else:
            return FreeCAD.Vector(V.X, V.Y, V.Z)

    def _processFeature(self, obj, shape, sub, pNum):
        """_processFeature(obj, shape, sub, pNum)...
        This function analyzes a shape and returns a three item tuple containing:
            working point,
            shape orientation/slope,
            shape category as face, edge, or vert."""
        p = None
        dYdX = None
        cat = sub[:4]
        PathLog.debug('sub-feature is {}'.format(cat))
        Ref = getattr(obj, 'Reference' + str(pNum))
        if cat == 'Face':
            BE = self._getBottomEdge(shape)
            if BE:
                self.bottomEdges.append(BE)
            # calculate slope of face
            V0 = shape.Vertexes[0]
            v1 = shape.CenterOfMass
            temp = FreeCAD.Vector(v1.x - V0.X, v1.y - V0.Y, 0.0)
            dYdX = self._normalizeVector(temp)

            # Determine normal vector for face
            norm = shape.normalAt(0.0, 0.0)
            # FreeCAD.Console.PrintMessage('{} normal {}.\n'.format(sub, norm))
            if norm.z != 0:
                msg = translate('PathSlot',
                    'The selected face is not oriented vertically:')
                FreeCAD.Console.PrintError(msg + ' {}.\n'.format(sub))
                return False

            if Ref == 'Center of Mass':
                comS = shape.CenterOfMass
                p = FreeCAD.Vector(comS.x, comS.y, 0.0)
            elif Ref == 'Center of BoundBox':
                comS = shape.BoundBox.Center
                p = FreeCAD.Vector(comS.x, comS.y, 0.0)
            elif Ref == 'Lowest Point':
                p = self._getLowestPoint(shape)
            elif Ref == 'Highest Point':
                p = self._getHighestPoint(shape)

        elif cat == 'Edge':
            # calculate slope between end vertexes
            v0 = shape.Edges[0].Vertexes[0]
            v1 = shape.Edges[0].Vertexes[1]
            temp = FreeCAD.Vector(v1.X - v0.X, v1.Y - v0.Y, 0.0)
            dYdX = self._normalizeVector(temp)

            if Ref == 'Center of Mass':
                comS = shape.CenterOfMass
                p = FreeCAD.Vector(comS.x, comS.y, 0.0)
            elif Ref == 'Center of BoundBox':
                comS = shape.BoundBox.Center
                p = FreeCAD.Vector(comS.x, comS.y, 0.0)
            elif Ref == 'Lowest Point':
                p = self._findLowestPointOnEdge(shape)
            elif Ref == 'Highest Point':
                p = self._findHighestPointOnEdge(shape)

        elif cat == 'Vert':
            V = shape.Vertexes[0]
            p = FreeCAD.Vector(V.X, V.Y, 0.0)

        if p:
            return (p, dYdX, cat)

        return False

    def _extendArcSlot(self, p1, p2, cent, begExt, endExt):
        """_extendArcSlot(p1, p2, cent, begExt, endExt)...
                This function extends an arc defined by two end points, p1 and p2, and the center.
        The arc is extended along the circumference with begExt and endExt values.
        The function returns the new end points as tuple (n1, n2) to replace p1 and p2."""
        cancel = True
        if not begExt and not endExt :
            return (p1, p2)

        n1 = p1
        n2 = p2

        # Create a chord of the right length, on XY plane, starting on x axis 
        def makeChord(rads):
            x = self.newRadius * math.cos(rads)
            y = self.newRadius * math.sin(rads)
            a = FreeCAD.Vector(self.newRadius, 0.0, 0.0)
            b = FreeCAD.Vector(x, y, 0.0)
            return Part.makeLine(a, b)


        # Convert extension to radians; make a generic chord ( line ) on XY plane from the x axis
        # rotate and shift into place so it has same vertices as the required arc extension
        # adjust rotation angle to provide +ve or -ve extension as needed 
        origin = FreeCAD.Vector(0.0, 0.0, 0.0)
        if begExt:
            ExtRadians = abs(begExt / self.newRadius)
            chord = makeChord(ExtRadians)

            beginRadians = self._getVectorAngle(p1.sub(self.arcCenter))
            if begExt < 0:
                beginRadians += 0  # negative Ext shortens slot so chord endpoint is slot start point 
            else :
                beginRadians -= 2*ExtRadians  # positive Ext lengthens slot so decrease start point angle
                
            # PathLog.debug('begExt angles are: {},  {}'.format(beginRadians, math.degrees(beginRadians)))

            chord.rotate(origin, FreeCAD.Vector(0, 0, 1), math.degrees(beginRadians))
            chord.translate(self.arcCenter)
            self._addDebugObject(chord, 'ExtendStart')
            
            v1 = chord.Vertexes[1]
            n1 = FreeCAD.Vector(v1.X, v1.Y, 0.0)
            
        if endExt:
            ExtRadians = abs(endExt / self.newRadius)
            chord = makeChord(ExtRadians)

            endRadians = self._getVectorAngle(p2.sub(self.arcCenter))
            if endExt > 0:
                endRadians += 0  # positive Ext lengthens slot so chord endpoint is good
            else :
                endRadians -= 2*ExtRadians  # negative Ext shortens slot so decrease end point angle
                
            # PathLog.debug('endExt angles are: {},  {}'.format(endRadians, math.degrees(endRadians)))

            chord.rotate(origin, FreeCAD.Vector(0, 0, 1), math.degrees(endRadians))
            chord.translate(self.arcCenter)
            self._addDebugObject(chord, 'ExtendEnd')
               
            v1 = chord.Vertexes[1]
            n2 = FreeCAD.Vector(v1.X, v1.Y, 0.0)

        return (n1, n2)

    def _makeOffsetArc(self, p1, p2, center, newRadius):
        """_makeOffsetArc(p1, p2, center, newRadius)...
        This function offsets an arc defined by endpoints, p1 and p2, and the center.
        New end points are returned at the radius passed by newRadius.
        The angle of the original arc is maintained."""
        n1 = p1.sub(center).normalize()*newRadius
        n2 = p2.sub(center).normalize()*newRadius
        return (n1.add(center), n2.add(center))

    def _extendLineSlot(self, p1, p2, begExt, endExt):
        """_extendLineSlot(p1, p2, begExt, endExt)...
        This function extends a line defined by endpoints, p1 and p2.
        The beginning is extended by begExt value and the end by endExt value."""
        if begExt:
            beg = p1.sub(p2)
            n1 = p1.add(beg.normalize()*begExt)
        else:
            n1 = p1
        if endExt:
          end = p2.sub(p1)
          n2 = p2.add(end.normalize()*endExt)
        else:
            n2 = p2
        return (n1, n2) 

    def _getOppMidPoints(self, same):
        """_getOppMidPoints(same)...
        Find mid-points between ends of equal, oppossing edges passed in tuple (edge1, edge2)."""
        com1 = same[0].CenterOfMass
        com2 = same[1].CenterOfMass
        p1 = FreeCAD.Vector(com1.x, com1.y, 0.0)
        p2 = FreeCAD.Vector(com2.x, com2.y, 0.0)
        return (p1, p2)

    def _isParallel(self, dYdX1, dYdX2):
        """Determine if two orientation vectors are parallel."""
        return (dYdX1.cross(dYdX2) == FreeCAD.Vector(0,0,0) )
 #       if dYdX1.add(dYdX2).Length == 0:
 #           return True
 #       if ((dYdX1.x + dYdX2.x) / 2.0 == dYdX1.x and
 #               (dYdX1.y + dYdX2.y) / 2.0 == dYdX1.y):
 #           return True
 #       return False

    def _makePerpendicular(self, p1, p2, length):
        """_makePerpendicular(p1, p2, length)...
        Using a line defined by p1 and p2, returns a perpendicular vector centered
        at the midpoint of the line, with length value."""
        line = Part.makeLine(p1, p2)
        midPnt = line.CenterOfMass

        halfDist = length / 2.0
        if self.dYdX1:
            half = FreeCAD.Vector(self.dYdX1.x, self.dYdX1.y, 0.0).multiply(halfDist)
            n1 = midPnt.add(half)
            n2 = midPnt.sub(half)
            return (n1, n2)
        elif self.dYdX2:
            half = FreeCAD.Vector(self.dYdX2.x, self.dYdX2.y, 0.0).multiply(halfDist)
            n1 = midPnt.add(half)
            n2 = midPnt.sub(half)
            return (n1, n2)
        else:
            toEnd = p2.sub(p1)
            perp = FreeCAD.Vector(-1 * toEnd.y, toEnd.x, 0.0)
            perp.normalize()
            perp.multiply(halfDist)
            n1 = midPnt.add(perp)
            n2 = midPnt.sub(perp)
            return (n1, n2)

    def _findLowestPointOnEdge(self, E):
        tol = 0.0000001
        zMin = E.BoundBox.ZMin
        # Test first vertex
        v = E.Vertexes[0]
        if abs(v.Z - zMin) < tol:
            return FreeCAD.Vector(v.X, v.Y, v.Z)
        # Test second vertex
        v = E.Vertexes[1]
        if abs(v.Z - zMin) < tol:
            return FreeCAD.Vector(v.X, v.Y, v.Z)
        # Test middle point of edge
        eMidLen = E.Length / 2.0
        eMidPnt = E.valueAt(E.getParameterByLength(eMidLen))
        if abs(eMidPnt.z - zMin) < tol:
            return eMidPnt
        if E.BoundBox.ZLength < 0.000000001:  # roughly horizontal edge
            return eMidPnt
        return self._findLowestEdgePoint(E)

    def _findLowestEdgePoint(self, E):
        zMin = E.BoundBox.ZMin
        eLen = E.Length
        L0 = 0
        L1 = eLen
        p0 = None
        p1 = None
        cnt = 0
        while L1 - L0 > 0.00001 and cnt < 2000:
            adj = (L1 - L0) * 0.1
            # Get points at L0 and L1 along edge
            p0 = E.valueAt(E.getParameterByLength(L0))
            p1 = E.valueAt(E.getParameterByLength(L1))
            # Adjust points based on proximity to target depth
            diff0 = p0.z - zMin
            diff1 = p1.z - zMin
            if diff0 < diff1:
                L1 -= adj
            elif diff0 > diff1:
                L0 += adj
            else:
                L0 += adj
                L1 -= adj
            cnt += 1
        midLen = (L0 + L1) / 2.0
        return E.valueAt(E.getParameterByLength(midLen))

    def _findHighestPointOnEdge(self, E):
        tol = 0.0000001
        zMax = E.BoundBox.ZMax
        # Test first vertex
        v = E.Vertexes[0]
        if abs(zMax - v.Z) < tol:
            return FreeCAD.Vector(v.X, v.Y, v.Z)
        # Test second vertex
        v = E.Vertexes[1]
        if abs(zMax - v.Z) < tol:
            return FreeCAD.Vector(v.X, v.Y, v.Z)
        # Test middle point of edge
        eMidLen = E.Length / 2.0
        eMidPnt = E.valueAt(E.getParameterByLength(eMidLen))
        if abs(zMax - eMidPnt.z) < tol:
            return eMidPnt
        if E.BoundBox.ZLength < 0.000000001:  # roughly horizontal edge
            return eMidPnt
        return self._findHighestEdgePoint(E)

    def _findHighestEdgePoint(self, E):
        zMax = E.BoundBox.ZMax
        eLen = E.Length
        L0 = 0
        L1 = eLen
        p0 = None
        p1 = None
        cnt = 0
        while L1 - L0 > 0.00001 and cnt < 2000:
            adj = (L1 - L0) * 0.1
            # Get points at L0 and L1 along edge
            p0 = E.valueAt(E.getParameterByLength(L0))
            p1 = E.valueAt(E.getParameterByLength(L1))
            # Adjust points based on proximity to target depth
            diff0 = zMax - p0.z
            diff1 = zMax - p1.z
            if diff0 < diff1:
                L1 -= adj
            elif diff0 > diff1:
                L0 += adj
            else:
                L0 += adj
                L1 -= adj
            cnt += 1
        midLen = (L0 + L1) / 2.0
        return E.valueAt(E.getParameterByLength(midLen))

    def _getVectorAngle(self, v):
        # Assumes Z value of vector is zero
        halfPi = math.pi / 2

        if v.y == 1 and v.x == 0:
            return halfPi
        if v.y == -1 and v.x == 0:
            return math.pi + halfPi
        if v.y == 0 and v.x == 1:
            return 0.0
        if v.y == 0 and v.x == -1:
            return math.pi

        x = abs(v.x)
        y = abs(v.y)
        rads = math.atan(y/x)
        if v.x > 0:
            if v.y > 0:
                return rads
            else:
                return (2 * math.pi) - rads
        if v.x < 0:
            if v.y > 0:
                return math.pi - rads
            else:
                return math.pi + rads

    def _getCutSidePoints(self, obj, v0, v1, a1, a2, b1, b2):
        ea1 = Part.makeLine(v0, a1)
        ea2 = Part.makeLine(a1, a2)
        ea3 = Part.makeLine(a2, v1)
        ea4 = Part.makeLine(v1, v0)
        boxA = Part.Face(Part.Wire([ea1, ea2, ea3, ea4]))
        cubeA = boxA.extrude(FreeCAD.Vector(0.0, 0.0, 1.0))
        cmnA = self.base.Shape.common(cubeA)
        eb1 = Part.makeLine(v0, b1)
        eb2 = Part.makeLine(b1, b2)
        eb3 = Part.makeLine(b2, v1)
        eb4 = Part.makeLine(v1, v0)
        boxB = Part.Face(Part.Wire([eb1, eb2, eb3, eb4]))
        cubeB = boxB.extrude(FreeCAD.Vector(0.0, 0.0, 1.0))
        cmnB = self.base.Shape.common(cubeB)
        if cmnA.Volume > cmnB.Volume:
            return (b1, b2)
        return (a1, a2)

    def _getBottomEdge(self, shape):
        EDGES = list()
        # Determine if selected face has a single bottom horizontal edge
        eCnt = len(shape.Edges)
        eZMin = shape.BoundBox.ZMin
        for ei in range(0, eCnt):
            E = shape.Edges[ei]
            if abs(E.BoundBox.ZMax - eZMin) < 0.00000001:
                EDGES.append(E)
        if len(EDGES) == 1:  # single bottom horiz. edge
            return EDGES[0]
        return False

    def _getVertFaceType(self, shape):
        wires = list()

        bottomEdge = self._getBottomEdge(shape)
        if bottomEdge:
            return ('Edge', bottomEdge)

        # Extract cross-section of face
        extFwd = (shape.BoundBox.ZLength * 2.2) + 10
        extShp = shape.extrude(FreeCAD.Vector(0.0, 0.0, extFwd))
        sliceZ = shape.BoundBox.ZMin + (extFwd / 2.0)
        slcs = extShp.slice(FreeCAD.Vector(0, 0, 1), sliceZ)
        for i in slcs:
            wires.append(i)
        if len(wires) > 0:
            if wires[0].isClosed():
                face = Part.Face(wires[0])
                if face.Area > 0:
                    face.translate(FreeCAD.Vector(0.0, 0.0, shape.BoundBox.ZMin - face.BoundBox.ZMin))
                    return ('Face', face)
            return ('Wire', wires[0])
        return False

    def _makeReference1Enumerations(self, sub, single=False):
        """Customize Reference1 enumerations based on feature type."""
        PathLog.debug('_makeReference1Enumerations()')
        cat = sub[:4]
        if single:
            if cat == 'Face':
                return ['Long Edge', 'Short Edge']
            elif cat == 'Edge':
                return ['Long Edge']
            elif cat == 'Vert':
                return ['Vertex']
        elif cat == 'Vert':
            return ['Vertex']

        return ['Center of Mass', 'Center of BoundBox',
               'Lowest Point', 'Highest Point']

    def _makeReference2Enumerations(self, sub):
        """Customize Reference2 enumerations based on feature type."""
        PathLog.debug('_makeReference2Enumerations()')
        cat = sub[:4]
        if cat == 'Vert':
            return ['Vertex']
        return ['Center of Mass', 'Center of BoundBox',
               'Lowest Point', 'Highest Point']

    def _lineCollisionCheck(self, obj, p1, p2):
        """Make simple circle with diameter of tool, at start point.
        Extrude it latterally along path.
        Extrude it vertically.
        Check for collision with model."""
        # Make path travel of tool as 3D solid.
        rad = self.tool.Diameter / 2.0

        def getPerp(p1, p2, dist):
            toEnd = p2.sub(p1)
            perp = FreeCAD.Vector(-1 * toEnd.y, toEnd.x, 0.0)
            if perp.x == 0 and perp.y == 0:
                return perp
            perp.normalize()
            perp.multiply(dist)
            return perp

        # Make first cylinder
        ce1 = Part.Wire(Part.makeCircle(rad, p1).Edges)
        C1 = Part.Face(ce1)
        zTrans = obj.FinalDepth.Value - C1.BoundBox.ZMin
        C1.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
        extFwd = obj.StartDepth.Value - obj.FinalDepth.Value
        extVect = FreeCAD.Vector(0.0, 0.0, extFwd)
        startShp = C1.extrude(extVect)

        if p2.sub(p1).Length > 0:
            # Make second cylinder
            ce2 = Part.Wire(Part.makeCircle(rad, p2).Edges)
            C2 = Part.Face(ce2)
            zTrans = obj.FinalDepth.Value - C2.BoundBox.ZMin
            C2.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            endShp = C2.extrude(extVect)

            # Make extruded rectangle to connect cylinders
            perp = getPerp(p1, p2, rad)
            v1 = p1.add(perp)
            v2 = p1.sub(perp)
            v3 = p2.sub(perp)
            v4 = p2.add(perp)
            e1 = Part.makeLine(v1, v2)
            e2 = Part.makeLine(v2, v3)
            e3 = Part.makeLine(v3, v4)
            e4 = Part.makeLine(v4, v1)
            edges = Part.__sortEdges__([e1, e2, e3, e4])
            rectFace = Part.Face(Part.Wire(edges))
            zTrans = obj.FinalDepth.Value - rectFace.BoundBox.ZMin
            rectFace.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            boxShp = rectFace.extrude(extVect)

            # Fuse two cylinders and box together
            part1 = startShp.fuse(boxShp)
            pathTravel = part1.fuse(endShp)
        else:
            pathTravel = startShp

        self._addDebugObject(pathTravel, 'PathTravel')

        # Check for collision with model
        try:
            cmn = self.base.Shape.common(pathTravel)
            if cmn.Volume > 0.000001:
                return True
        except Exception:
            PathLog.debug('Failed to complete path collision check.')

        return False

    def _arcCollisionCheck(self, obj, p1, p2, arcCenter, arcRadius):
        """Make simple circle with diameter of tool, at start and end points.
        Make arch face between circles. Fuse and extrude it vertically.
        Check for collision with model."""
        # Make path travel of tool as 3D solid.
        if hasattr(self.tool.Diameter, 'Value'):
            rad = self.tool.Diameter.Value / 2.0
        else:
            rad = self.tool.Diameter / 2.0
        extFwd = obj.StartDepth.Value - obj.FinalDepth.Value
        extVect = FreeCAD.Vector(0.0, 0.0, extFwd)

        if self.isArc == 1:
            # full circular slot
            # make outer circle
            oCircle = Part.makeCircle(arcRadius + rad, arcCenter)
            oWire = Part.Wire(oCircle.Edges[0])
            outer = Part.Face(oWire)
            # make inner circle
            iRadius = arcRadius - rad
            if iRadius > 0:
                iCircle = Part.makeCircle(iRadius, arcCenter)
                iWire = Part.Wire(iCircle.Edges[0])
                inner = Part.Face(iWire)
                # Cut outer with inner
                path = outer.cut(inner)
            else:
                path = outer
            zTrans = obj.FinalDepth.Value - path.BoundBox.ZMin
            path.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            pathTravel = path.extrude(extVect)
        else:
            # arc slot
            # Make first cylinder
            ce1 = Part.Wire(Part.makeCircle(rad, p1).Edges)
            C1 = Part.Face(ce1)
            zTrans = obj.FinalDepth.Value - C1.BoundBox.ZMin
            C1.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            startShp = C1.extrude(extVect)
            # self._addDebugObject(startShp, 'StartCyl')

            # Make second cylinder
            ce2 = Part.Wire(Part.makeCircle(rad, p2).Edges)
            C2 = Part.Face(ce2)
            zTrans = obj.FinalDepth.Value - C2.BoundBox.ZMin
            C2.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            endShp = C2.extrude(extVect)
            # self._addDebugObject(endShp, 'EndCyl')

            # Make wire with inside and outside arcs, and lines on ends.
            # Convert wire to face, then extrude

            # verify offset does not force radius < 0
            newRadius = arcRadius - rad
            # PathLog.debug('arcRadius, newRadius: {}, {}'.format(arcRadius, newRadius))
            if newRadius <= 0:
                msg = translate('PathSlot',
                        'Current offset value produces negative radius.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False
            else:
                (pA, pB) = self._makeOffsetArc(p1, p2, arcCenter, newRadius)
                arc_inside = Arcs.arcFrom2Pts(pA, pB, arcCenter)

            # Arc 2 - outside
            # verify offset does not force radius < 0
            newRadius = arcRadius + rad
            # PathLog.debug('arcRadius, newRadius: {}, {}'.format(arcRadius, newRadius))
            if newRadius <= 0:
                msg = translate('PathSlot',
                        'Current offset value produces negative radius.')
                FreeCAD.Console.PrintError(msg + '\n')
                return False
            else:
                (pC, pD) = self._makeOffsetArc(p1, p2, arcCenter, newRadius)
                arc_outside = Arcs.arcFrom2Pts(pC, pD, arcCenter)

            # Make end lines to connect arcs
            vA = arc_inside.Vertexes[0]
            vB = arc_inside.Vertexes[1]
            vC = arc_outside.Vertexes[1]
            vD = arc_outside.Vertexes[0]
            pa = FreeCAD.Vector(vA.X, vA.Y, 0.0)
            pb = FreeCAD.Vector(vB.X, vB.Y, 0.0)
            pc = FreeCAD.Vector(vC.X, vC.Y, 0.0)
            pd = FreeCAD.Vector(vD.X, vD.Y, 0.0)

            # Make closed arch face and extrude
            e1 = Part.makeLine(pb, pc)
            e2 = Part.makeLine(pd, pa)
            edges = Part.__sortEdges__([arc_inside, e1, arc_outside, e2])
            rectFace = Part.Face(Part.Wire(edges))
            zTrans = obj.FinalDepth.Value - rectFace.BoundBox.ZMin
            rectFace.translate(FreeCAD.Vector(0.0, 0.0, zTrans))
            boxShp = rectFace.extrude(extVect)
            # self._addDebugObject(boxShp, 'ArcBox')

            # Fuse two cylinders and box together
            part1 = startShp.fuse(boxShp)
            pathTravel = part1.fuse(endShp)

        self._addDebugObject(pathTravel, 'PathTravel')

        # Check for collision with model
        try:
            cmn = self.base.Shape.common(pathTravel)
            if cmn.Volume > 0.000001:
                print ("volume=",cmn.Volume)
                return True
        except Exception:
            PathLog.debug('Failed to complete path collision check.')

        return False

    def _addDebugObject(self, objShape, objName):
        if self.showDebugObjects:
            do = FreeCAD.ActiveDocument.addObject('Part::Feature', 'tmp_' + objName)
            do.Shape = objShape
            do.purgeTouched()
            self.tmpGrp.addObject(do)
# Eclass


def SetupProperties():
    ''' SetupProperties() ... Return list of properties required for operation.'''
    return [tup[1] for tup in ObjectSlot.opPropertyDefinitions(False)]


def Create(name, obj=None):
    '''Create(name) ... Creates and returns a Slot operation.'''
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectSlot(obj, name)
    return obj



# ***************************************************************************
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Helper functions that are used by IFC importer and exporter."""
import sys
import math
import six

import FreeCAD
import Arch
import ArchIFC

from draftutils.messages import _msg, _wrn

PREDEFINED_RGB = {"black": (0, 0, 0),
                  "red": (1.0, 0, 0),
                  "green": (0, 1.0, 0),
                  "blue": (0, 0, 1.0),
                  "yellow": (1.0, 1.0, 0),
                  "magenta": (1.0, 0, 1.0),
                  "cyan": (0, 1.0, 1.0),
                  "white": (1.0, 1.0, 1.0)}


DEBUG_prod_repr = False
DEBUG_prod_colors = False


def decode(filename, utf=False):
    """Turn unicode into strings, only for Python 2."""
    if six.PY2 and isinstance(filename, six.text_type):
        # This is a workaround since ifcopenshell 0.6 currently
        # can't handle unicode filenames
        encoding = "utf8" if utf else sys.getfilesystemencoding()
        filename = filename.encode(encoding)
    return filename


def dd2dms(dd):
    """Convert decimal degrees to degrees, minutes, seconds.

    Used in export.
    """
    sign = 1 if dd >= 0 else -1
    dd = abs(dd)
    minutes, seconds = divmod(dd * 3600, 60)
    degrees, minutes = divmod(minutes, 60)

    if dd < 0:
        degrees = -degrees

    return (int(degrees) * sign,
            int(minutes) * sign,
            int(seconds) * sign)


def dms2dd(degrees, minutes, seconds, milliseconds=0):
    """Convert degrees, minutes, seconds to decimal degrees.

    Used in import.
    """
    dd = float(degrees) + float(minutes)/60 + float(seconds)/3600
    return dd


class ProjectImporter:
    """A helper class to create an Arch Project object."""

    def __init__(self, file, objects):
        self.file = file
        self.objects = objects

    def execute(self):
        self.project = self.file.by_type("IfcProject")[0]
        self.object = Arch.makeProject()
        self.objects[self.project.id()] = self.object
        self.setAttributes()
        self.setComplexAttributes()

    def setAttributes(self):
        for prop in self.object.PropertiesList:
            if hasattr(self.project, prop) and getattr(self.project, prop):
                setattr(self.object, prop, getattr(self.project, prop))

    def setComplexAttributes(self):
        try:
            mapConversion = self.project.RepresentationContexts[0].HasCoordinateOperation[0]

            data = self.extractTargetCRSData(mapConversion.TargetCRS)
            data.update(self.extractMapConversionData(mapConversion))
            # TODO: review and refactor this piece of code.
            # Calling a method from a class is a bit strange;
            # this class should be derived from that class to inherit
            # this method; otherwise a simple function (not tied to a class)
            # should be used.
            ArchIFC.IfcRoot.setObjIfcComplexAttributeValue(self, self.object, "RepresentationContexts", data)
        except Exception:
            # This scenario occurs validly in IFC2X3,
            # as the mapConversion does not exist
            return

    def extractTargetCRSData(self, targetCRS):
        mappings = {
            "name": "Name",
            "description": "Description",
            "geodetic_datum": "GeodeticDatum",
            "vertical_datum": "VerticalDatum",
            "map_projection": "MapProjection",
            "map_zone": "MapZone"
        }
        data = {}
        for attributeName, ifcName in mappings.items():
            data[attributeName] = str(getattr(targetCRS, ifcName))

        if targetCRS.MapUnit.Prefix:
            data["map_unit"] = targetCRS.MapUnit.Prefix.title() + targetCRS.MapUnit.Name.lower()
        else:
            data["map_unit"] = targetCRS.MapUnit.Name.title()

        return data

    def extractMapConversionData(self, mapConversion):
        mappings = {
            "eastings": "Eastings",
            "northings": "Northings",
            "orthogonal_height": "OrthogonalHeight",
            "x_axis_abscissa": "XAxisAbscissa",
            "x_axis_ordinate": "XAxisOrdinate",
            "scale": "Scale"
        }
        data = {}
        for attributeName, ifcName in mappings.items():
            data[attributeName] = str(getattr(mapConversion, ifcName))

        data["true_north"] = str(self.calculateTrueNorthAngle(mapConversion.XAxisAbscissa,
                                                              mapConversion.XAxisOrdinate))
        return data

    def calculateTrueNorthAngle(self, x, y):
        return round(math.degrees(math.atan2(y, x)) - 90, 6)


def buildRelProductsAnnotations(ifcfile, root_element='IfcProduct'):
    """Build the products and annotations relation table."""
    products = ifcfile.by_type(root_element)

    annotations = ifcfile.by_type("IfcAnnotation")
    tp = []
    for product in products:
        if product.is_a("IfcGrid") and (product not in annotations):
            annotations.append(product)
        elif product not in annotations:
            tp.append(product)

    # remove any leftover annotations from products
    products = sorted(tp, key=lambda prod: prod.id())

    return products, annotations


def buildRelProductRepresentation(ifcfile):
    """Build the product/representations relation table."""
    if DEBUG_prod_repr:
        _msg(32 * "-")
        _msg("Product-representation table")

    prodrepr = dict()

    i = 1
    for p in ifcfile.by_type("IfcProduct"):
        if hasattr(p, "Representation") and p.Representation:
            if DEBUG_prod_repr:
                _msg("{}: {}, {}, '{}'".format(i, p.id(),
                                               p.is_a(), p.Name))

            for it in p.Representation.Representations:
                for it1 in it.Items:
                    prodrepr.setdefault(p.id(), []).append(it1.id())
                    if it1.is_a("IfcBooleanResult"):
                        prodrepr.setdefault(p.id(), []).append(it1.FirstOperand.id())
                    elif it.Items[0].is_a("IfcMappedItem"):
                        prodrepr.setdefault(p.id(), []).append(it1.MappingSource.MappedRepresentation.id())
                        if it1.MappingSource.MappedRepresentation.is_a("IfcShapeRepresentation"):
                            for it2 in it1.MappingSource.MappedRepresentation.Items:
                                prodrepr.setdefault(p.id(), []).append(it2.id())
            i += 1
    return prodrepr


def buildRelAdditions(ifcfile):
    """Build the additions relation table."""
    additions = {}  # { host:[child,...], ... }

    for r in ifcfile.by_type("IfcRelContainedInSpatialStructure"):
        additions.setdefault(r.RelatingStructure.id(), []).extend([e.id() for e in r.RelatedElements])
    for r in ifcfile.by_type("IfcRelAggregates"):
        additions.setdefault(r.RelatingObject.id(), []).extend([e.id() for e in r.RelatedObjects])

    return additions


def buildRelGroups(ifcfile):
    """Build the groups relation table."""
    groups = {}  # { host:[child,...], ... }     # used in structural IFC

    for r in ifcfile.by_type("IfcRelAssignsToGroup"):
        groups.setdefault(r.RelatingGroup.id(), []).extend([e.id() for e in r.RelatedObjects])

    return groups


def buildRelSubtractions(ifcfile):
    """Build the subtractions relation table."""
    subtractions = []  # [ [opening,host], ... ]

    for r in ifcfile.by_type("IfcRelVoidsElement"):
        subtractions.append([r.RelatedOpeningElement.id(), r.RelatingBuildingElement.id()])

    return subtractions


def buildRelMattable(ifcfile):
    """Build the mattable relation table."""
    mattable = {}  # { objid:matid }

    for r in ifcfile.by_type("IfcRelAssociatesMaterial"):
        for o in r.RelatedObjects:
            if r.RelatingMaterial.is_a("IfcMaterial"):
                mattable[o.id()] = r.RelatingMaterial.id()
            elif r.RelatingMaterial.is_a("IfcMaterialLayer"):
                mattable[o.id()] = r.RelatingMaterial.Material.id()
            elif r.RelatingMaterial.is_a("IfcMaterialLayerSet"):
                mattable[o.id()] = r.RelatingMaterial.MaterialLayers[0].Material.id()
            elif r.RelatingMaterial.is_a("IfcMaterialLayerSetUsage"):
                mattable[o.id()] = r.RelatingMaterial.ForLayerSet.MaterialLayers[0].Material.id()

    return mattable


# Color relation tables.
# Products can have a color, materials can have a color,
# and products can have a material.
# Colors for material assigned to a product, and color of the product itself
# can be different
def buildRelColors(ifcfile, prodrepr):
    """Build the colors relation table.

    Returns all IfcStyledItem colors, material and product colors.

    Returns
    -------
    dict
        A dictionary with `{id: (r,g,b), ...}` values.
    """
    colors = {}  # { id:(r,g,b) }
    style_material_id = {}  # { style_entity_id: material_id) }

    style_color_rgb = {}  # { style_entity_id: (r,g,b) }
    for r in ifcfile.by_type("IfcStyledItem"):
        if r.Styles and r.Styles[0].is_a("IfcPresentationStyleAssignment"):
            for style1 in r.Styles[0].Styles:
                if style1.is_a("IfcSurfaceStyle"):
                    for style2 in style1.Styles:
                        if style2.is_a("IfcSurfaceStyleRendering"):
                            if style2.SurfaceColour:
                                c = style2.SurfaceColour
                                style_color_rgb[r.id()] = (c.Red,
                                                           c.Green,
                                                           c.Blue)

        # Nova
        # FIXME: style_entity_id = { style_entity_id: product_id } not material_id ???
        # see https://forum.freecadweb.org/viewtopic.php?f=39&t=37940&start=10#p329491
        # last code change in these color code https://github.com/FreeCAD/FreeCAD/commit/2d1f6ab1
        '''
        if r.Item:
            # print(r.id())
            # print(r.Item)  # IfcRepresentationItem or IfcShapeRepresentation
            for p in prodrepr.keys():
                if r.Item.id() in prodrepr[p]:
                    style_material_id[r.id()] = p
                    # print(p)
                    # print(ifcfile[p])  # product
        '''

    # A much faster version for Nova style_material_id with product_ids
    # no material colors, Nova ifc files often do not have materials at all
    for p in prodrepr.keys():
        # print("\n")
        # print(ifcfile[p])  # IfcProduct
        # print(ifcfile[p].Representation)  # IfcProductDefinitionShape
        # print(ifcfile[p].Representation.Representations[0])  # IfcShapeRepresentation
        # print(ifcfile[p].Representation.Representations[0].Items[0])  # IfcRepresentationItem
        # print(ifcfile[p].Representation.Representations[0].Items[0].StyledByItem[0])  # IfcStyledItem
        # print(ifcfile[p].Representation.Representations[0].Items[0].StyledByItem[0].id())
        # print(p)
        representation_item = ifcfile[p].Representation.Representations[0].Items[0]
        if hasattr(representation_item, "StyledByItem") and representation_item.StyledByItem:
            style_material_id[representation_item.StyledByItem[0].id()] = p

    # Allplan, ArchiCAD
    for m in ifcfile.by_type("IfcMaterialDefinitionRepresentation"):
        for it in m.Representations:
            if it.Items:
                style_material_id[it.Items[0].id()] = m.RepresentedMaterial.id()

    # create colors out of style_color_rgb and style_material_id
    for k in style_material_id:
        if k in style_color_rgb:
            colors[style_material_id[k]] = style_color_rgb[k]

    return colors


def buildRelProductColors(ifcfile, prodrepr):
    """Build the colors relation table from a product.

    Returns
    -------
    dict
        A dictionary with `{id: (r,g,b), ...}` values.
    """
    if DEBUG_prod_repr:
        _msg(32 * "-")
        _msg("Product-color table")

    colors = dict()
    i = 0

    for p in prodrepr.keys():
        # Representation item, see `IfcRepresentationItem` documentation.
        # All kinds of geometric or topological representation items
        # `IfcExtrudedAreaSolid`, `IfcMappedItem`, `IfcFacetedBrep`,
        # `IfcBooleanResult`, `IfcBooleanClippingResult`, etc.
        _body = ifcfile[p].Representation.Representations[0]
        repr_item = _body.Items[0]

        if DEBUG_prod_colors:
            _msg("{}: {}, {}, '{}', rep_item {}".format(i, ifcfile[p].id(),
                                                        ifcfile[p].is_a(),
                                                        ifcfile[p].Name,
                                                        repr_item))
        # Get the geometric representations which have a presentation style.
        # All representation items have the inverse attribute `StyledByItem`
        # for this.
        # There will be geometric representations which do not have
        # a presentation style so `StyledByItem` will be empty.
        if repr_item.StyledByItem:
            if DEBUG_prod_colors:
                _msg("  StyledByItem -> {}".format(repr_item.StyledByItem))
            # it has to be a `IfcStyledItem`, no check needed
            styled_item = repr_item.StyledByItem[0]

            # Write into colors table if a `IfcStyledItem` exists
            # for this product, write `None` if something goes wrong
            # or if the ifc file has errors and thus no valid color
            # is returned
            colors[p] = getColorFromStyledItem(styled_item)

        i += 1
    return colors


def buildRelMaterialColors(ifcfile, prodrepr):
    # not implemented
    pass


def getColorFromProduct(product):

    if product.Representation:
        for rep in product.Representation.Representations:
            for item in rep.Items:
                for style in item.StyledByItem:
                    color = getColorFromStyledItem(style)
                    if color:
                        return color

def getColorFromMaterial(material):

    if material.HasRepresentation:
        rep = material.HasRepresentation[0]
        if hasattr(rep,"Representations") and rep.Representations:
            rep = rep.Representations[0]
            if rep.is_a("IfcStyledRepresentation"):
                return getColorFromStyledItem(rep)
    return None


def getColorFromStyledItem(styled_item):
    """Get color from the IfcStyledItem.

    Returns
    -------
    float, float, float, int
        A tuple with the red, green, blue, and transparency values.
        If the `IfcStyledItem` is a `IfcDraughtingPreDefinedColour`
        the transparency is set to 0.
        The first three values range from 0 to 1.0, while the transparency
        varies from 0 to 100.

    None
        Return `None` if `styled_item` is not of type `'IfcStyledItem'`
        or if there is any other problem getting a color.
    """

    if styled_item.is_a("IfcStyledRepresentation"):
        styled_item = styled_item.Items[0]

    if not styled_item.is_a("IfcStyledItem"):
        return None

    rgb_color = None
    transparency = None
    col = None

    # The `IfcStyledItem` holds presentation style information for products,
    # either explicitly for an `IfcGeometricRepresentationItem` being part of
    # an `IfcShapeRepresentation` assigned to a product, or by assigning
    # presentation information to `IfcMaterial` being assigned
    # as other representation for a product.

    # In current IFC release (IFC2x3) only one presentation style
    # assignment shall be assigned.
    # In IFC4 `IfcPresentationStyleAssignment` is deprecated
    # In IFC4 multiple styles are assigned to style in 'IfcStyleItem' instead

    # print(ifcfile[p])
    # print(styled_item)
    # print(styled_item.Styles)
    if len(styled_item.Styles) == 0:
        # IN IFC2x3, only one element in `Styles` should be available.
        _wrn("No 'Style' in 'IfcStyleItem', do nothing.")
        # ca 100x in 210_King_Merged.ifc
        # Empty styles, #4952778=IfcStyledItem(#4952779,(),$)
        # this is an error in the IFC file in my opinion
    else:
        # never seen an ifc with more than one Styles in IfcStyledItem
        # the above seams to only apply for IFC2x3, IFC4 can have them
        # see https://forum.freecadweb.org/viewtopic.php?f=39&t=33560&p=437056#p437056

        # Get the `IfcPresentationStyleAssignment`, there should only be one,
        if styled_item.Styles[0].is_a('IfcPresentationStyleAssignment'):
            assign_style = styled_item.Styles[0]
        else:
            # `IfcPresentationStyleAssignment` is deprecated in IFC4,
            # in favor of `IfcStyleAssignmentSelect`
            assign_style = styled_item
        # print(assign_style)  # IfcPresentationStyleAssignment

        # `IfcPresentationStyleAssignment` can hold various kinds and counts
        # of styles, see `IfcPresentationStyleSelect`
        if assign_style.Styles[0].is_a("IfcSurfaceStyle"):
            _style = assign_style.Styles[0]
            # Schependomlaan and Nova and others
            # `IfcSurfaceStyleRendering`
            # print(_style.Styles[0])
            # `IfcColourRgb`
            rgb_color = _style.Styles[0].SurfaceColour
            # print(rgb_color)
            if (_style.Styles[0].is_a('IfcSurfaceStyleShading')
                    and hasattr(_style.Styles[0], 'Transparency')
                    and _style.Styles[0].Transparency):
                transparency = _style.Styles[0].Transparency * 100
        elif assign_style.Styles[0].is_a("IfcCurveStyle"):
            if (len(assign_style.Styles) == 2
                    and assign_style.Styles[1].is_a("IfcSurfaceStyle")):
                # Allplan, new IFC export started in 2017
                # `IfcDraughtingPreDefinedColour`
                # print(assign_style.Styles[0].CurveColour)
                # TODO: check this; on index 1, is this what we need?!
                rgb_color = assign_style.Styles[1].Styles[0].SurfaceColour
                # print(rgb_color)
            else:
                # 2x Annotations in 210_King_Merged.ifc
                # print(ifcfile[p])
                # print(assign_style.Styles[0])
                # print(assign_style.Styles[0].CurveColour)
                rgb_color = assign_style.Styles[0].CurveColour

    if rgb_color:
        if rgb_color.is_a('IfcDraughtingPreDefinedColour'):
            if DEBUG_prod_colors:
                _msg("  '{}'= ".format(rgb_color.Name))

            col = predefined_to_rgb(rgb_color)

            if col:
                col = col + (0, )
        else:
            col = (rgb_color.Red,
                   rgb_color.Green,
                   rgb_color.Blue,
                   int(transparency) if transparency else 0)
    else:
        col = None

    if DEBUG_prod_colors:
        _msg("  {}".format(col))

    return col


def predefined_to_rgb(rgb_color):
    """Transform a predefined color name to its [r, g, b] representation.

    TODO: at the moment it doesn't handle 'by layer'.
    See: `IfcDraughtingPreDefinedColour` and `IfcPresentationLayerWithStyle`.
    """
    name = rgb_color.Name.lower()
    if name not in PREDEFINED_RGB:
        _wrn("Color name not in 'IfcDraughtingPreDefinedColour'.")

        if name == 'by layer':
            _wrn("'IfcDraughtingPreDefinedColour' set 'by layer'; "
                 "currently not handled, set to 'None'.")
        return None

    return PREDEFINED_RGB[name]

# ************************************************************************************************
# property related methods
def buildRelProperties(ifcfile):
    """
    Builds and returns a dictionary of {object:[properties]} from an IFC file
    """

    # this method no longer used by the importer module
    # but this relation table might be useful anyway for other purposes

    properties = {}  # { objid : { psetid : [propertyid, ... ], ... }, ... }
    for r in ifcfile.by_type("IfcRelDefinesByProperties"):
        for obj in r.RelatedObjects:
            if not obj.id() in properties:
                properties[obj.id()] = {}
            psets = {}
            props = []
            if r.RelatingPropertyDefinition.is_a("IfcPropertySet"):
                props.extend([prop.id() for prop in r.RelatingPropertyDefinition.HasProperties])
                psets[r.RelatingPropertyDefinition.id()] = props
                properties[obj.id()].update(psets)
    return properties


def getIfcPropertySets(ifcfile, pid):
    """Returns a dictionary of {pset_id:[prop_id, prop_id...]} for an IFC object"""

    # get psets for this pid
    psets = {}
    for rel in ifcfile[pid].IsDefinedBy:
        # the following if condition is needed in IFC2x3 only
        # https://forum.freecadweb.org/viewtopic.php?f=39&t=37892#p322884
        if rel.is_a('IfcRelDefinesByProperties'):
            props = []
            if rel.RelatingPropertyDefinition.is_a("IfcPropertySet"):
                props.extend([prop.id() for prop in rel.RelatingPropertyDefinition.HasProperties])
                psets[rel.RelatingPropertyDefinition.id()] = props
    return psets


def getIfcProperties(ifcfile, pid, psets, d):
    """builds valid property values for FreeCAD"""

    for pset in psets.keys():
        # print("reading pset: ",pset)
        psetname = ifcfile[pset].Name
        if six.PY2:
            psetname = psetname.encode("utf8")
        for prop in psets[pset]:
            e = ifcfile[prop]
            pname = e.Name
            if six.PY2:
                pname = pname.encode("utf8")
            if e.is_a("IfcPropertySingleValue"):
                if e.NominalValue:
                    ptype = e.NominalValue.is_a()
                    if ptype in ['IfcLabel','IfcText','IfcIdentifier','IfcDescriptiveMeasure']:
                        pvalue = e.NominalValue.wrappedValue
                        if six.PY2:
                            pvalue = pvalue.encode("utf8")
                    else:
                        pvalue = str(e.NominalValue.wrappedValue)
                    if hasattr(e.NominalValue,'Unit'):
                        if e.NominalValue.Unit:
                            pvalue += e.NominalValue.Unit
                    d[pname+";;"+psetname] = ptype+";;"+pvalue
                # print("adding property: ",pname,ptype,pvalue," pset ",psetname)
    return d


def getIfcPsetPoperties(ifcfile, pid):
    """ directly build the property table from pid and ifcfile for FreeCAD"""

    return getIfcProperties(ifcfile, pid, getIfcPropertySets(ifcfile, pid), {})


def getUnit(unit):
    """Get the unit multiplier for different decimal prefixes.

    Only for when the unit is METRE.
    When no Prefix is provided, return 1000, that is, mm x 1000 = metre.
    For other cases, return 1.0.
    """
    if unit.Name == "METRE":
        if unit.Prefix == "KILO":
            return 1000000.0
        elif unit.Prefix == "HECTO":
            return 100000.0
        elif unit.Prefix == "DECA":
            return 10000.0
        elif not unit.Prefix:
            return 1000.0
        elif unit.Prefix == "DECI":
            return 100.0
        elif unit.Prefix == "CENTI":
            return 10.0
    return 1.0


def getScaling(ifcfile):
    """Return a scaling factor from the IFC file; units to mm."""
    ua = ifcfile.by_type("IfcUnitAssignment")

    if not ua:
        return 1.0

    ua = ua[0]
    for u in ua.Units:
        if u.UnitType == "LENGTHUNIT":
            if u.is_a("IfcConversionBasedUnit"):
                f = getUnit(u.ConversionFactor.UnitComponent)
                return f * u.ConversionFactor.ValueComponent.wrappedValue
            elif u.is_a("IfcSIUnit") or u.is_a("IfcUnit"):
                return getUnit(u)
    return 1.0


def getRotation(entity):
    """returns a FreeCAD rotation from an IfcProduct with a IfcMappedItem representation"""
    try:
        u = FreeCAD.Vector(entity.Axis1.DirectionRatios)
        v = FreeCAD.Vector(entity.Axis2.DirectionRatios)
        w = FreeCAD.Vector(entity.Axis3.DirectionRatios)
    except AttributeError:
        return FreeCAD.Rotation()
    import WorkingPlane
    p = WorkingPlane.plane(u=u, v=v, w=w)
    return p.getRotation().Rotation


def getPlacement(entity,scaling=1000):
    """returns a placement from the given entity"""

    if not entity:
        return None
    import DraftVecUtils
    pl = None
    if entity.is_a("IfcAxis2Placement3D"):
        x = getVector(entity.RefDirection,scaling)
        z = getVector(entity.Axis,scaling)
        if x and z:
            y = z.cross(x)
            m = DraftVecUtils.getPlaneRotation(x,y,z)
            pl = FreeCAD.Placement(m)
        else:
            pl = FreeCAD.Placement()
        loc = getVector(entity.Location,scaling)
        if loc:
            pl.move(loc)
    elif entity.is_a("IfcAxis2Placement2D"):
        _wrn("not implemented IfcAxis2Placement2D, ", end="")
    elif entity.is_a("IfcLocalPlacement"):
        pl = getPlacement(entity.PlacementRelTo,1)  # original placement
        relpl = getPlacement(entity.RelativePlacement,1)  # relative transf
        if pl and relpl:
            pl = pl.multiply(relpl)
        elif relpl:
            pl = relpl
    elif entity.is_a("IfcCartesianPoint"):
        loc = getVector(entity,scaling)
        pl = FreeCAD.Placement()
        pl.move(loc)
    if pl:
        pl.Base = FreeCAD.Vector(pl.Base).multiply(scaling)
    return pl


def getVector(entity,scaling=1000):
    """returns a vector from the given entity"""

    if not entity:
        return None
    v = None
    if entity.is_a("IfcDirection"):
        if len(entity.DirectionRatios) == 3:
            v = FreeCAD.Vector(tuple(entity.DirectionRatios))
        else:
            v = FreeCAD.Vector(tuple(entity.DirectionRatios+[0]))
    elif entity.is_a("IfcCartesianPoint"):
        if len(entity.Coordinates) == 3:
            v = FreeCAD.Vector(tuple(entity.Coordinates))
        else:
            v = FreeCAD.Vector(tuple(entity.Coordinates+[0]))
    # if v:
    #     v.multiply(scaling)
    return v


def get2DShape(representation,scaling=1000):
    """Returns a shape from a 2D IfcShapeRepresentation"""

    import Part
    import DraftVecUtils
    import Draft

    def getPolyline(ent):
        pts = []
        for p in ent.Points:
            c = p.Coordinates
            c = FreeCAD.Vector(c[0],c[1],c[2] if len(c) > 2 else 0)
            c.multiply(scaling)
            pts.append(c)
        return Part.makePolygon(pts)

    def getRectangle(ent):
        return Part.makePlane(ent.XDim,ent.YDim)

    def getLine(ent):
        pts = []
        p1 = getVector(ent.Pnt)
        p1.multiply(scaling)
        pts.append(p1)
        p2 = getVector(ent.Dir)
        p2.multiply(scaling)
        p2 = p1.add(p2)
        pts.append(p2)
        return Part.makePolygon(pts)

    def getCircle(ent):
        c = ent.Position.Location.Coordinates
        c = FreeCAD.Vector(c[0],c[1],c[2] if len(c) > 2 else 0)
        c.multiply(scaling)
        r = ent.Radius*scaling
        return Part.makeCircle(r,c)

    def getCurveSet(ent):
        result = []
        if ent.is_a() in ["IfcGeometricCurveSet","IfcGeometricSet"]:
            elts = ent.Elements
        elif ent.is_a() in ["IfcLine","IfcPolyline","IfcCircle","IfcTrimmedCurve","IfcRectangleProfileDef"]:
            elts = [ent]
        else:
            print("getCurveSet: unhandled entity: ", ent)
            return []

        for el in elts:
            if el.is_a("IfcPolyline"):
                result.append(getPolyline(el))
            elif el.is_a("IfcRectangleProfileDef"):
                result.append(getRectangle(el))
            elif el.is_a("IfcLine"):
                result.append(getLine(el))
            elif el.is_a("IfcCircle"):
                result.append(getCircle(el))
            elif el.is_a("IfcTrimmedCurve"):
                base = el.BasisCurve
                t1 = el.Trim1[0].wrappedValue
                t2 = el.Trim2[0].wrappedValue
                if not el.SenseAgreement:
                    t1,t2 = t2,t1
                if base.is_a("IfcPolyline"):
                    bc = getPolyline(base)
                    result.append(bc)
                elif base.is_a("IfcCircle"):
                    bc = getCircle(base)
                    e = Part.ArcOfCircle(bc.Curve,math.radians(t1),math.radians(t2)).toShape()
                    d = base.Position.RefDirection.DirectionRatios
                    v = FreeCAD.Vector(d[0],d[1],d[2] if len(d) > 2 else 0)
                    a = -DraftVecUtils.angle(v)
                    e.rotate(bc.Curve.Center,FreeCAD.Vector(0,0,1),math.degrees(a))
                    result.append(e)
            elif el.is_a("IfcCompositeCurve"):
                for base in el.Segments:
                    if base.ParentCurve.is_a("IfcPolyline"):
                        bc = getPolyline(base.ParentCurve)
                        result.append(bc)
                    elif base.ParentCurve.is_a("IfcCircle"):
                        bc = getCircle(base.ParentCurve)
                        e = Part.ArcOfCircle(bc.Curve,math.radians(t1),math.radians(t2)).toShape()
                        d = base.Position.RefDirection.DirectionRatios
                        v = FreeCAD.Vector(d[0],d[1],d[2] if len(d) > 2 else 0)
                        a = -DraftVecUtils.angle(v)
                        e.rotate(bc.Curve.Center,FreeCAD.Vector(0,0,1),math.degrees(a))
                        result.append(e)
            elif el.is_a("IfcIndexedPolyCurve"):
                coords = el.Points.CoordList
                def index2points(segment):
                    pts = []
                    for i in segment.wrappedValue:
                        c = coords[i-1]
                        c = FreeCAD.Vector(c[0],c[1],c[2] if len(c) > 2 else 0)
                        c.multiply(scaling)
                        pts.append(c)
                    return pts

                for s in el.Segments:
                    if s.is_a("IfcLineIndex"):
                        result.append(Part.makePolygon(index2points(s)))
                    elif s.is_a("IfcArcIndex"):
                        [p1, p2, p3] = index2points(s)
                        result.append(Part.Arc(p1, p2, p3))
                    else:
                        raise RuntimeError("Illegal IfcIndexedPolyCurve segment")
            else:
                print("getCurveSet: unhandled element: ", el)

        return result

    result = []
    if representation.is_a("IfcShapeRepresentation"):
        for item in representation.Items:
            if item.is_a() in ["IfcGeometricCurveSet","IfcGeometricSet"]:
                result = getCurveSet(item)
            elif item.is_a("IfcMappedItem"):
                preresult = get2DShape(item.MappingSource.MappedRepresentation,scaling)
                pla = getPlacement(item.MappingSource.MappingOrigin,scaling)
                rot = getRotation(item.MappingTarget)
                if pla:
                    if rot.Angle:
                        pla.Rotation = rot
                    for r in preresult:
                        # r.Placement = pla
                        result.append(r)
                else:
                    result = preresult
            elif item.is_a("IfcTextLiteral"):
                pl = getPlacement(item.Placement, scaling)
                if pl:
                    t = Draft.makeText([item.Literal], point=pl.Base)
                    return [t]  # dirty hack... Object creation should not be done here
    elif representation.is_a() in ["IfcPolyline","IfcCircle","IfcTrimmedCurve","IfcRectangleProfileDef"]:
        result = getCurveSet(representation)
    return result


def getProfileCenterPoint(sweptsolid):
    """returns the center point of the profile of an extrusion"""
    v = FreeCAD.Vector(0,0,0)
    if hasattr(sweptsolid,"SweptArea"):
        profile = get2DShape(sweptsolid.SweptArea)
        if profile:
            profile = profile[0]
            if hasattr(profile,"CenterOfMass"):
                v = profile.CenterOfMass
            elif hasattr(profile,"BoundBox"):
                v = profile.BoundBox.Center
    if hasattr(sweptsolid,"Position"):
        pos = getPlacement(sweptsolid.Position)
        v = pos.multVec(v)
    return v


def isRectangle(verts):
    """returns True if the given 4 vertices form a rectangle"""
    if len(verts) != 4:
        return False
    v1 = verts[1].sub(verts[0])
    v2 = verts[2].sub(verts[1])
    v3 = verts[3].sub(verts[2])
    v4 = verts[0].sub(verts[3])
    if abs(v2.getAngle(v1)-math.pi/2) > 0.01:
        return False
    if abs(v3.getAngle(v2)-math.pi/2) > 0.01:
        return False
    if abs(v4.getAngle(v3)-math.pi/2) > 0.01:
        return False
    return True


def createFromProperties(propsets,ifcfile,parametrics):

    """
    Creates a FreeCAD parametric object from a set of properties.
    """

    obj = None
    sets = []
    appset = None
    guiset = None
    for pset in propsets.keys():
        if ifcfile[pset].Name == "FreeCADPropertySet":
            appset = {}
            for pid in propsets[pset]:
                p = ifcfile[pid]
                appset[p.Name] = p.NominalValue.wrappedValue
        elif ifcfile[pset].Name == "FreeCADGuiPropertySet":
            guiset = {}
            for pid in propsets[pset]:
                p = ifcfile[pid]
                guiset[p.Name] = p.NominalValue.wrappedValue
    if appset:
        oname = None
        otype = None
        if "FreeCADType" in appset.keys():
            if "FreeCADName" in appset.keys():
                obj = FreeCAD.ActiveDocument.addObject(appset["FreeCADType"],appset["FreeCADName"])
                if "FreeCADAppObject" in appset:
                    mod,cla = appset["FreeCADAppObject"].split(".")
                    if "'" in mod:
                        mod = mod.split("'")[-1]
                    if "'" in cla:
                        cla = cla.split("'")[0]
                    import importlib
                    mod = importlib.import_module(mod)
                    getattr(mod,cla)(obj)
                sets.append(("App",appset))
                if FreeCAD.GuiUp:
                    if guiset:
                        if "FreeCADGuiObject" in guiset:
                            mod,cla = guiset["FreeCADGuiObject"].split(".")
                            if "'" in mod:
                                mod = mod.split("'")[-1]
                            if "'" in cla:
                                cla = cla.split("'")[0]
                            import importlib
                            mod = importlib.import_module(mod)
                            getattr(mod,cla)(obj.ViewObject)
                        sets.append(("Gui",guiset))
    if obj and sets:
        for realm,pset in sets:
            if realm == "App":
                target = obj
            else:
                target = obj.ViewObject
            for key,val in pset.items():
                if key.startswith("FreeCAD_") or key.startswith("FreeCADGui_"):
                    name = key.split("_")[1]
                    if name in target.PropertiesList:
                        if not target.getEditorMode(name):
                            ptype = target.getTypeIdOfProperty(name)
                            if ptype in ["App::PropertyString","App::PropertyEnumeration","App::PropertyInteger","App::PropertyFloat"]:
                                setattr(target,name,val)
                            elif ptype in ["App::PropertyLength","App::PropertyDistance"]:
                                setattr(target,name,val*1000)
                            elif ptype == "App::PropertyBool":
                                if val in [".T.",True]:
                                    setattr(target,name,True)
                                else:
                                    setattr(target,name,False)
                            elif ptype == "App::PropertyVector":
                                setattr(target,name,FreeCAD.Vector([float(s) for s in val.split("(")[1].strip(")").split(",")]))
                            elif ptype == "App::PropertyArea":
                                setattr(target,name,val*1000000)
                            elif ptype == "App::PropertyPlacement":
                                data = val.split("[")[1].strip("]").split("(")
                                data = [data[1].split(")")[0],data[2].strip(")")]
                                v = FreeCAD.Vector([float(s) for s in data[0].split(",")])
                                r = FreeCAD.Rotation(*[float(s) for s in data[1].split(",")])
                                setattr(target,name,FreeCAD.Placement(v,r))
                            elif ptype == "App::PropertyLink":
                                link = val.split("_")[1]
                                parametrics.append([target,name,link])
                            else:
                                print("Unhandled FreeCAD property:",name," of type:",ptype)
    return obj,parametrics


def applyColorDict(doc,colordict=None):

    """applies the contents of a color dict to the objects in the given doc.
    If no colordict is given, the doc Meta property is searched for a "colordict" entry."""

    if not colordict:
        if "colordict" in doc.Meta:
            import json
            colordict = json.loads(doc.Meta["colordict"])
    if colordict:
        for obj in doc.Objects:
            if obj.Name in colordict:
                color = colordict[obj.Name]
                if hasattr(obj.ViewObject,"ShapeColor"):
                    obj.ViewObject.ShapeColor = tuple(color[0:3])
                if hasattr(obj.ViewObject,"Transparency") and (len(color) >= 4):
                    obj.ViewObject.Transparency = color[3]
    else:
        print("No valid color dict to apply")


def getParents(ifcobj):

    """finds the parent entities of an IFC entity"""

    parentlist = []
    if hasattr(ifcobj,"ContainedInStructure"):
        for rel in ifcobj.ContainedInStructure:
            parentlist.append(rel.RelatingStructure)
    elif hasattr(ifcobj,"Decomposes"):
        for rel in ifcobj.Decomposes:
            if rel.is_a("IfcRelAggregates"):
                parentlist.append(rel.RelatingObject)
    return parentlist



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to scale shapes."""
## @package scale
# \ingroup draftfuctions
# \brief Provides functions to scale shapes.

## \addtogroup draftfuctions
# @{
import FreeCAD as App
import DraftVecUtils
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils
import draftfunctions.join as join
import draftmake.make_copy as make_copy


def scale(objectslist, scale=App.Vector(1,1,1),
          center=App.Vector(0,0,0), copy=False):
    """scale(objects, scale, [center], copy)
    
    Scales the objects contained in objects (that can be a list of objects or 
    an object) of the given  around given center. 

    Parameters
    ----------
    objectlist : list

    scale : Base.Vector
        Scale factors defined by a given vector (in X, Y, Z directions).

    objectlist : Base.Vector
        Center of the scale operation.

    copy : bool
        If copy is True, the actual objects are not scaled, but copies
        are created instead. 

    Return
    ----------
    The objects (or their copies) are returned.
    """
    if not isinstance(objectslist, list):
        objectslist = [objectslist]
    newobjlist = []
    for obj in objectslist:
        if copy:
            newobj = make_copy.make_copy(obj)
        else:
            newobj = obj
        if hasattr(obj,'Shape'):
            scaled_shape = obj.Shape.copy()
            m = App.Matrix()
            m.move(center.negative())
            m.scale(scale.x,scale.y,scale.z)
            m.move(center)
            scaled_shape = scaled_shape.transformGeometry(m)
        if utils.get_type(obj) == "Rectangle":
            p = []
            for v in scaled_shape.Vertexes:
                p.append(v.Point)
            pl = obj.Placement.copy()
            pl.Base = p[0]
            diag = p[2].sub(p[0])
            bb = p[1].sub(p[0])
            bh = p[3].sub(p[0])
            nb = DraftVecUtils.project(diag,bb)
            nh = DraftVecUtils.project(diag,bh)
            if obj.Length < 0: l = -nb.Length
            else: l = nb.Length
            if obj.Height < 0: h = -nh.Length
            else: h = nh.Length
            newobj.Length = l
            newobj.Height = h
            tr = p[0].sub(obj.Shape.Vertexes[0].Point) # unused?
            newobj.Placement = pl
        elif utils.get_type(obj) == "Wire" or utils.get_type(obj) == "BSpline":
            for index, point in enumerate(newobj.Points):
                scale_vertex(newobj, index, scale, center)
        elif hasattr(obj,'Shape'):
            newobj.Shape = scaled_shape
        elif hasattr(obj,"Position"):
            d = obj.Position.sub(center)
            newobj.Position = center.add(App.Vector(d.x * scale.x,
                                                    d.y * scale.y,
                                                    d.z * scale.z))
        elif hasattr(obj,"Placement"):
            d = obj.Placement.Base.sub(center)
            newobj.Placement.Base = center.add(App.Vector(d.x * scale.x,
                                                    d.y * scale.y,
                                                    d.z * scale.z))
            if hasattr(obj,"Height"):
                obj.setExpression('Height', None)
                obj.Height = obj.Height * scale.y
            if hasattr(obj,"Width"):
                obj.setExpression('Width', None)
                obj.Width = obj.Width * scale.x
            if hasattr(obj,"XSize"):
                obj.setExpression('XSize', None)
                obj.XSize = obj.XSize * scale.x
            if hasattr(obj,"YSize"):
                obj.setExpression('YSize', None)
                obj.YSize = obj.YSize * scale.y
        if obj.ViewObject and hasattr(obj.ViewObject,"FontSize"):
            obj.ViewObject.FontSize = obj.ViewObject.FontSize * scale.y
                
                
        if copy:
            gui_utils.format_object(newobj,obj)
        newobjlist.append(newobj)
    if copy and utils.get_param("selectBaseObjects",False):
        gui_utils.select(objectslist)
    else:
        gui_utils.select(newobjlist)
    if len(newobjlist) == 1: return newobjlist[0]
    return newobjlist


#   Following functions are needed for SubObjects modifiers
#   implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire)


def scale_vertex(obj, vertex_index, scale, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    points = obj.Points
    points[vertex_index] = obj.Placement.inverse().multVec(
        scaleVectorFromCenter(
            obj.Placement.multVec(points[vertex_index]),
            scale, center))
    obj.Points = points


scaleVertex = scale_vertex


def scale_vector_from_center(vector, scale, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    return vector.sub(center).scale(scale.x, scale.y, scale.z).add(center)


scaleVectorFromCenter = scale_vector_from_center


def scale_edge(obj, edge_index, scale, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    scaleVertex(obj, edge_index, scale, center)
    if utils.isClosedEdge(edge_index, obj):
        scaleVertex(obj, 0, scale, center)
    else:
        scaleVertex(obj, edge_index+1, scale, center)


scaleEdge = scale_edge


def copy_scaled_edge(obj, edge_index, scale, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    import Part
    vertex1 = scaleVectorFromCenter(
        obj.Placement.multVec(obj.Points[edge_index]),
        scale, center)
    if utils.isClosedEdge(edge_index, obj):
        vertex2 = scaleVectorFromCenter(
            obj.Placement.multVec(obj.Points[0]),
            scale, center)
    else:
        vertex2 = scaleVectorFromCenter(
            obj.Placement.multVec(obj.Points[edge_index+1]),
            scale, center)
    return Part.makeLine(vertex1, vertex2)


copyScaledEdge = copy_scaled_edge


def copy_scaled_edges(arguments):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    copied_edges = []
    for argument in arguments:
        copied_edges.append(copyScaledEdge(argument[0], argument[1],
            argument[2], argument[3]))
    join.join_wires(copied_edges)


copyScaledEdges = copy_scaled_edges

## @}



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2018 Benjamin Alterauge (ageeye)                        *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create PointArray objects.

The copies will be placed along a list of points defined by a sketch,
a `Part::Compound`, or a `Draft Block`.
"""
## @package make_pointarray
# \ingroup draftmake
# \brief Provides functions to create PointArray objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftutils.messages import _msg, _err
from draftutils.translate import translate
from draftobjects.pointarray import PointArray

if App.GuiUp:
    from draftviewproviders.view_array import ViewProviderDraftArray
    from draftviewproviders.view_draftlink import ViewProviderDraftLink


def make_point_array(base_object, point_object, extra=None, use_link=True):
    """Make a Draft PointArray object.

    Distribute copies of a `base_object` in the points
    defined by `point_object`.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    point_object: Part::Feature or str
        An object that is a type of container for holding points.
        This object must have one of the following properties `Geometry`,
        `Links`, or `Components`, which themselves must contain objects
        with `X`, `Y`, and `Z` properties.

        This object could be:

        - A `Sketcher::SketchObject`, as it has a `Geometry` property.
          The sketch can contain different elements but it must contain
          at least one `Part::GeomPoint`.

        - A `Part::Compound`, as it has a `Links` property. The compound
          can contain different elements but it must contain at least
          one object that has `X`, `Y`, and `Z` properties,
          like a `Draft Point` or a `Part::Vertex`.

        - A `Draft Block`, as it has a `Components` property. This `Block`
          behaves essentially the same as a `Part::Compound`. It must
          contain at least a point or vertex object.

    extra: Base::Placement, Base::Vector3, or Base::Rotation, optional
        It defaults to `None`.
        If it is provided, it is an additional placement that is applied
        to each copy of the array.
        The input could be a full placement, just a vector indicating
        the additional translation, or just a rotation.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'PointArray'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.
    """
    _name = "make_point_array"
    utils.print_header(_name, "Point array")

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft","No active document. Aborting."))
        return None

    if isinstance(base_object, str):
        base_object_str = base_object

    found, base_object = utils.find_object(base_object, doc)
    if not found:
        _msg("base_object: {}".format(base_object_str))
        _err(translate("draft","Wrong input: object not in document."))
        return None

    _msg("base_object: {}".format(base_object.Label))

    if isinstance(point_object, str):
        point_object_str = point_object

    found, point_object = utils.find_object(point_object, doc)
    if not found:
        _msg("point_object: {}".format(point_object_str))
        _err(translate("draft","Wrong input: object not in document."))
        return None

    _msg("point_object: {}".format(point_object.Label))
    if (not hasattr(point_object, "Geometry")
            and not hasattr(point_object, "Links")
            and not hasattr(point_object, "Components")):
        _err(translate("draft","Wrong input: point object doesn't have 'Geometry', 'Links', or 'Components'."))
        return None

    _msg("extra: {}".format(extra))
    if not extra:
        extra = App.Placement()
    try:
        utils.type_check([(extra, (App.Placement,
                                   App.Vector,
                                   App.Rotation))],
                         name=_name)
    except TypeError:
        _err(translate("draft","Wrong input: must be a placement, a vector, or a rotation."))
        return None

    # Convert the vector or rotation to a full placement
    if isinstance(extra, App.Vector):
        extra = App.Placement(extra, App.Rotation())
    elif isinstance(extra, App.Rotation):
        extra = App.Placement(App.Vector(), extra)

    if use_link:
        # The PointArray class must be called in this special way
        # to make it a LinkArray
        new_obj = doc.addObject("Part::FeaturePython", "PointArray",
                                PointArray(None), None, True)
    else:
        new_obj = doc.addObject("Part::FeaturePython", "PointArray")
        PointArray(new_obj)

    new_obj.Base = base_object
    new_obj.PointObject = point_object
    new_obj.ExtraPlacement = extra

    if App.GuiUp:
        if use_link:
            ViewProviderDraftLink(new_obj.ViewObject)
        else:
            ViewProviderDraftArray(new_obj.ViewObject)
            gui_utils.format_object(new_obj, new_obj.Base)

            if hasattr(new_obj.Base.ViewObject, "DiffuseColor"):
                if len(new_obj.Base.ViewObject.DiffuseColor) > 1:
                    new_obj.ViewObject.Proxy.resetColors(new_obj.ViewObject)

        new_obj.Base.ViewObject.hide()
        gui_utils.select(new_obj)

    return new_obj


def makePointArray(base, ptlst):
    """Create PointArray. DEPRECATED. Use 'make_point_array'."""
    utils.use_instead('make_point_array')

    return make_point_array(base, ptlst)

## @}



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2013 WandererFan <wandererfan@gmail.com>                *
# *   Copyright (c) 2019 Zheng, Lei (realthunder)<realthunder.dev@gmail.com>*
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the PathArray object.

The copies will be placed along a path like a polyline, spline, or bezier
curve, and along the selected subelements.

To Do
-----
The `'PathSubelements'` property must be changed in type, as it does not need
to be an `App::PropertyLinkSubList`.
A `LinkSubList` is to select multiple subelements (edges) from multiple
objects. However, since we need to select a `'Path Object'` already,
which is a single object, the subelements that we can choose must belong
to this `'Path Object'` only.

Therefore, the correct property that must be used is `App::PropertyLinkSub`.
Then in the property editor we will be unable to select more than one object
thus preventing errors of the subelements not matching the `'PathObject'`.

In fact, both `'PathObject'` and `'PathSubelements'`
could be handled with a single `App::PropertyLinkSub` property,
as this property can be used to select a single object,
or a single object with its subelements.

In the future, we could migrate the properties, or outright break
compatibility with older objects by changing both properties
`'PathObject'` and `'PathSubelements'`.

An alternative to this would be to use a single `App::PropertyLinkSubList`.
This would allow us to build PathArrays on multiple objects and multiple
subelements (edges) of those objects at the same time. However, to do this,
the logic in `execute` would have to be changed to account for multiple
objects. Therefore, the first solution is simpler, that is, using
a single property of type `App::PropertyLinkSub`.
"""
## @package patharray
# \ingroup draftobjects
# \brief Provides the object code for the PathArray object.

import FreeCAD as App
import DraftVecUtils
import lazy_loader.lazy_loader as lz

from draftutils.messages import _msg, _wrn, _err
from draftutils.translate import translate
def QT_TRANSLATE_NOOP(ctx,txt): return txt
from draftobjects.base import DraftObject
from draftobjects.draftlink import DraftLink

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")

## \addtogroup draftobjects
# @{


class PathArray(DraftLink):
    """The Draft Path Array object.

    The object distributes copies of an object along a path like a polyline,
    spline, or bezier curve.

    Attributes
    ----------
    Align: bool
        It defaults to `False`.
        It sets whether the object will be specially aligned to the path.

    AlignMode: str
        It defaults to `'Original'`.
        Indicates the type of alignment that will be calculated when
        `Align` is `True`.

        `'Original'` mode is the historic `'Align'` for old (v0.18) documents.
        It is not really the Fernat alignment. It uses the normal parameter
        from `getNormal` (or the default) as a constant, it does not calculate
        curve normal.
        `X` is curve tangent, `Y` is normal parameter, `Z` is the cross product
        `X` x `Y`.

        `'Tangent'` mode is similar to `Original`, but includes a pre-rotation
        (in execute) to align the `Base` object's `X` to `TangentVector`,
        then `X` follows curve tangent, normal input parameter
        is the Z component.

        If `ForceVertical` is `True`, the normal parameter from `getNormal`
        is ignored, and `X` is curve tangent, `Z` is `VerticalVector`,
        and `Y` is the cross product `X` x `Z`.

        `'Frenet'` mode orients the copies to a coordinate system
        along the path.
        `X` is tangent to curve, `Y` is curve normal, `Z` is curve binormal.
        If normal cannot be computed, for example, in a straight line,
        the default is used.

    ForceVertical: bool
        It defaults to `False`.
        If it is `True`, and `AlignMode` is `'Original'` or `'Tangent'`,
        it will use the vector in `VerticalVector` as the `Z` axis.
    """

    def __init__(self, obj):
        super(PathArray, self).__init__(obj, "PathArray")

    def attach(self, obj):
        """Set up the properties when the object is attached.

        Note: we don't exactly know why the properties are added
        in the `attach` method. They should probably be added in the `__init__`
        method. Maybe this is related to the link behavior of this class.

        For PathLinkArray, DraftLink.attach creates the link to the Base.

        Realthunder: before the big merge, there was only the attach() method
        in the view object proxy, not the object proxy.
        I added that to allow the proxy to override the C++ view provider
        type. The view provider type is normally determined by object's
        C++ API getViewProviderName(), and cannot be overridden by the proxy.
        I introduced the attach() method in proxy to allow the core
        to attach the proxy before creating the C++ view provider.
        """
        self.set_properties(obj)
        super(PathArray, self).attach(obj)

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        if not obj:
            return

        if hasattr(obj, "PropertiesList"):
            properties = obj.PropertiesList
        else:
            properties = []

        self.set_general_properties(obj, properties)
        self.set_align_properties(obj, properties)

    def set_general_properties(self, obj, properties):
        """Set general properties only if they don't exist."""
        if "Base" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","The base object that will be duplicated")
            obj.addProperty("App::PropertyLinkGlobal",
                            "Base",
                            "Objects",
                            _tip)
            obj.Base = None

        if "PathObject" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","The object along which the copies will be distributed. It must contain 'Edges'.")
            obj.addProperty("App::PropertyLinkGlobal",
                            "PathObject",
                            "Objects",
                            _tip)
            obj.PathObject = None

        # TODO: the 'PathSubelements' property must be changed,
        # as it does not need to be an 'App::PropertyLinkSubList'.
        #
        # In fact, both 'PathObject' and 'PathSubelements'
        # could be handled with a single 'App::PropertyLinkSub' property,
        # as this property can be used to select a single object,
        # or a single object with its subelements.
        if "PathSubelements" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","List of connected edges in the 'Path Object'.\nIf these are present, the copies will be created along these subelements only.\nLeave this property empty to create copies along the entire 'Path Object'.")
            obj.addProperty("App::PropertyLinkSubListGlobal",
                            "PathSubelements",
                            "Objects",
                            _tip)
            obj.PathSubelements = []

        if "Count" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Number of copies to create")
            obj.addProperty("App::PropertyInteger",
                            "Count",
                            "Objects",
                            _tip)
            obj.Count = 4

        if self.use_link and "ExpandArray" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Show the individual array elements (only for Link arrays)")
            obj.addProperty("App::PropertyBool",
                            "ExpandArray",
                            "Objects",
                            _tip)
            obj.ExpandArray = False
            obj.setPropertyStatus('Shape', 'Transient')

    def set_align_properties(self, obj, properties):
        """Set general properties only if they don't exist."""
        if "ExtraTranslation" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Additional translation that will be applied to each copy.\nThis is useful to adjust for the difference between shape centre and shape reference point.")
            obj.addProperty("App::PropertyVectorDistance",
                            "ExtraTranslation",
                            "Alignment",
                            _tip)
            obj.ExtraTranslation = App.Vector(0, 0, 0)

        if "TangentVector" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Alignment vector for 'Tangent' mode")
            obj.addProperty("App::PropertyVector",
                            "TangentVector",
                            "Alignment",
                            _tip)
            obj.TangentVector = App.Vector(1, 0, 0)

        if "ForceVertical" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Force use of 'Vertical Vector' as local Z direction when using 'Original' or 'Tangent' alignment mode")
            obj.addProperty("App::PropertyBool",
                            "ForceVertical",
                            "Alignment",
                            _tip)
            obj.ForceVertical = False

        if "VerticalVector" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Direction of the local Z axis when 'Force Vertical' is true")
            obj.addProperty("App::PropertyVector",
                            "VerticalVector",
                            "Alignment",
                            _tip)
            obj.VerticalVector = App.Vector(0, 0, 1)

        if "AlignMode" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Method to orient the copies along the path.\n- Original: X is curve tangent, Y is normal, and Z is the cross product.\n- Frenet: aligns the object following the local coordinate system along the path.\n- Tangent: similar to 'Original' but the local X axis is pre-aligned to 'Tangent Vector'.\n\nTo get better results with 'Original' or 'Tangent' you may have to set 'Force Vertical' to true.")
            obj.addProperty("App::PropertyEnumeration",
                            "AlignMode",
                            "Alignment",
                            _tip)
            obj.AlignMode = ['Original', 'Frenet', 'Tangent']
            obj.AlignMode = 'Original'

        # The Align property must be attached after other align properties
        # so that onChanged works properly
        if "Align" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Orient the copies along the path depending on the 'Align Mode'.\nOtherwise the copies will have the same orientation as the original Base object.")
            obj.addProperty("App::PropertyBool",
                            "Align",
                            "Alignment",
                            _tip)
            obj.Align = False

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super(PathArray, self).linkSetup(obj)
        obj.configLinkProperty(ElementCount='Count')

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        if not obj.Base or not obj.PathObject:
            return

        # placement of entire PathArray object
        array_placement = obj.Placement

        w = self.get_wires(obj.PathObject, obj.PathSubelements)
        if not w:
            _err(obj.PathObject.Label
                 + translate("draft",", path object doesn't have 'Edges'."))
            return

        base_rotation = obj.Base.Shape.Placement.Rotation
        final_rotation = base_rotation

        if (obj.Align and obj.AlignMode == "Tangent"
                and hasattr(obj, "TangentVector")):
            Xaxis = App.Vector(1.0, 0.0, 0.0)  # default TangentVector

            if not DraftVecUtils.equals(Xaxis, obj.TangentVector):
                # make rotation from TangentVector to X
                pre_rotation = App.Rotation(obj.TangentVector, Xaxis)
                final_rotation = base_rotation.multiply(pre_rotation)

        copy_placements = placements_on_path(final_rotation,
                                             w, obj.Count,
                                             obj.ExtraTranslation,
                                             obj.Align, obj.AlignMode,
                                             obj.ForceVertical,
                                             obj.VerticalVector)

        return super(PathArray, self).buildShape(obj,
                                                 array_placement,
                                                 copy_placements)

    def get_wires(self, path_object, subelements):
        """Get wires from the path object."""
        if subelements:
            w = self.get_wire_from_subelements(subelements)
        elif (hasattr(path_object.Shape, 'Wires')
              and path_object.Shape.Wires):
            w = path_object.Shape.Wires[0]
        elif path_object.Shape.Edges:
            w = Part.Wire(path_object.Shape.Edges)
        else:
            w = None
        return w

    def get_wire_from_subelements(self, subelements):
        """Make a wire from the path object subelements."""
        sl = []
        for sub in subelements:
            edgeNames = sub[1]
            for n in edgeNames:
                e = sub[0].Shape.getElement(n)
                sl.append(e)
        return Part.Wire(sl)

    def onChanged(self, obj, prop):
        """Execute when a property is changed."""
        super(PathArray, self).onChanged(obj, prop)
        self.show_and_hide(obj, prop)

    def show_and_hide(self, obj, prop):
        """Show and hide the properties depending on the touched property."""
        # The minus sign removes the Hidden property (show)
        if prop == "Align":
            if obj.Align:
                for pr in ("AlignMode", "ForceVertical", "VerticalVector",
                           "TangentVector"):
                    obj.setPropertyStatus(pr, "-Hidden")
            else:
                for pr in ("AlignMode", "ForceVertical", "VerticalVector",
                           "TangentVector"):
                    obj.setPropertyStatus(pr, "Hidden")

        if prop == "AlignMode":
            if obj.AlignMode == "Original":
                for pr in ("ForceVertical", "VerticalVector"):
                    obj.setPropertyStatus(pr, "-Hidden")

                obj.setPropertyStatus("TangentVector", "Hidden")

            elif obj.AlignMode == "Frenet":
                for pr in ("ForceVertical", "VerticalVector",
                           "TangentVector"):
                    obj.setPropertyStatus(pr, "Hidden")

            elif obj.AlignMode == "Tangent":
                for pr in ("ForceVertical", "VerticalVector",
                           "TangentVector"):
                    obj.setPropertyStatus(pr, "-Hidden")

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored.

        Add properties that don't exist.
        """
        super(PathArray, self).migrate_attributes(obj)
        self.set_properties(obj)
        self.migrate_properties_0v19(obj)

        if self.use_link:
            self.linkSetup(obj)
        else:
            obj.setPropertyStatus('Shape', '-Transient')

        if obj.Shape.isNull():
            if getattr(obj, 'PlacementList', None):
                self.buildShape(obj, obj.Placement, obj.PlacementList)
            else:
                self.execute(obj)

    def migrate_properties_0v19(self, obj):
        """Migrate properties of this class, not from the parent class."""
        properties = obj.PropertiesList

        if "PathObj" in properties:
            obj.PathObject = obj.PathObj
            obj.removeProperty("PathObj")
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","'PathObj' property will be migrated to 'PathObject'"))

        if "PathSubs" in properties:
            obj.PathSubelements = obj.PathSubs
            obj.removeProperty("PathSubs")
            _info = "'PathSubs' property will be migrated to 'PathSubelements'"
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","'PathObj' property will be migrated to 'PathObject'"))

        if "Xlate" in properties:
            obj.ExtraTranslation = obj.Xlate
            obj.removeProperty("Xlate")
            _info = "'Xlate' property will be migrated to 'ExtraTranslation'"
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","'PathObj' property will be migrated to 'PathObject'"))


# Alias for compatibility with v0.18 and earlier
_PathArray = PathArray


def placements_on_path(shapeRotation, pathwire, count, xlate, align,
                       mode='Original', forceNormal=False,
                       normalOverride=None):
    """Calculate the placements of a shape along a given path.

    Each copy will be distributed evenly.
    """
    closedpath = DraftGeomUtils.isReallyClosed(pathwire)
    normal = DraftGeomUtils.get_normal(pathwire)
    # for backward compatibility with previous getNormal implementation
    if normal is None:
        normal = App.Vector(0, 0, 1)

    if forceNormal and normalOverride:
        normal = normalOverride

    path = Part.__sortEdges__(pathwire.Edges)
    ends = []
    cdist = 0

    for e in path:  # find cumulative edge end distance
        cdist += e.Length
        ends.append(cdist)

    placements = []

    # place the start shape
    pt = path[0].Vertexes[0].Point
    _place = calculate_placement(shapeRotation,
                                 path[0], 0, pt, xlate, align, normal,
                                 mode, forceNormal)
    placements.append(_place)

    # closed path doesn't need shape on last vertex
    if not closedpath:
        # place the end shape
        pt = path[-1].Vertexes[-1].Point
        _place = calculate_placement(shapeRotation,
                                     path[-1], path[-1].Length,
                                     pt, xlate, align, normal,
                                     mode, forceNormal)
        placements.append(_place)

    if count < 3:
        return placements

    # place the middle shapes
    if closedpath:
        stop = count
    else:
        stop = count - 1
    step = float(cdist) / stop
    remains = 0
    travel = step
    for i in range(1, stop):
        # which edge in path should contain this shape?
        # avoids problems with float math travel > ends[-1]
        iend = len(ends) - 1

        for j in range(0, len(ends)):
            if travel <= ends[j]:
                iend = j
                break

        # place shape at proper spot on proper edge
        remains = ends[iend] - travel
        offset = path[iend].Length - remains
        pt = path[iend].valueAt(get_parameter_from_v0(path[iend], offset))

        _place = calculate_placement(shapeRotation,
                                     path[iend], offset,
                                     pt, xlate, align, normal,
                                     mode, forceNormal)
        placements.append(_place)

        travel += step

    return placements


calculatePlacementsOnPath = placements_on_path


def calculate_placement(globalRotation,
                        edge, offset, RefPt, xlate, align, normal=None,
                        mode='Original', overrideNormal=False):
    """Orient shape to a local coordinate system (tangent, normal, binormal).

    Orient shape at parameter offset, normally length.

    http://en.wikipedia.org/wiki/Euler_angles (previous version)
    http://en.wikipedia.org/wiki/Quaternions
    """
    # Start with a null Placement so the translation goes to the right place.
    # Then apply the global orientation.
    placement = App.Placement()
    placement.Rotation = globalRotation

    placement.move(RefPt + xlate)
    if not align:
        return placement

    nullv = App.Vector(0, 0, 0)
    defNormal = App.Vector(0.0, 0.0, 1.0)
    if normal:
        defNormal = normal

    try:
        t = edge.tangentAt(get_parameter_from_v0(edge, offset))
        t.normalize()
    except:
        _wrn(translate("draft","Cannot calculate path tangent. Copy not aligned."))
        return placement

    if mode in ('Original', 'Tangent'):
        if normal is None:
            n = defNormal
        else:
            n = normal
            n.normalize()

        try:
            b = t.cross(n)
            b.normalize()
        except Exception:
            # weird special case, tangent and normal parallel
            b = nullv
            _wrn(translate("draft","Tangent and normal are parallel. Copy not aligned."))
            return placement

        if overrideNormal:
            priority = "XZY"
            newRot = App.Rotation(t, b, n, priority)  # t/x, b/y, n/z
        else:
            # must follow X, try to follow Z, Y is what it is
            priority = "XZY"
            newRot = App.Rotation(t, n, b, priority)

    elif mode == 'Frenet':
        try:
            n = edge.normalAt(get_parameter_from_v0(edge, offset))
            n.normalize()
        except App.Base.FreeCADError:  # no/infinite normals here
            n = defNormal
            _msg(translate("draft","Cannot calculate path normal, using default."))

        try:
            b = t.cross(n)
            b.normalize()
        except Exception:
            b = nullv
            _wrn(translate("draft","Cannot calculate path binormal. Copy not aligned."))
            return placement

        priority = "XZY"
        newRot = App.Rotation(t, n, b, priority)  # t/x, n/y, b/z
    else:
        _msg(translate("draft","AlignMode {} is not implemented").format(mode))
        return placement

    # Have valid tangent, normal, binormal
    newGRot = newRot.multiply(globalRotation)

    placement.Rotation = newGRot
    return placement


calculatePlacement = calculate_placement


def get_parameter_from_v0(edge, offset):
    """Return parameter at distance offset from edge.Vertexes[0].

    sb method in Part.TopoShapeEdge???
    """
    lpt = edge.valueAt(edge.getParameterByLength(0))
    vpt = edge.Vertexes[0].Point

    if not DraftVecUtils.equals(vpt, lpt):
        # this edge is flipped
        length = edge.Length - offset
    else:
        # this edge is right way around
        length = offset

    return edge.getParameterByLength(length)


getParameterFromV0 = get_parameter_from_v0

## @}



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCADGui
import PathScripts.PathPreferences as PathPreferences
import PySide

# Qt translation handling
def translate(context, text, disambig=None):
    return PySide.QtCore.QCoreApplication.translate(context, text, disambig)

class AdvancedPreferencesPage:
    def __init__(self, parent=None):
        self.form = FreeCADGui.PySideUic.loadUi(':preferences/Advanced.ui')
        self.form.WarningSuppressAllSpeeds.stateChanged.connect(self.updateSelection)
        self.form.EnableAdvancedOCLFeatures.stateChanged.connect(self.updateSelection)

    def saveSettings(self):
        PathPreferences.setPreferencesAdvanced(
                self.form.EnableAdvancedOCLFeatures.isChecked(),
                self.form.WarningSuppressAllSpeeds.isChecked(),
                self.form.WarningSuppressRapidSpeeds.isChecked(),
                self.form.WarningSuppressSelectionMode.isChecked(),
                self.form.WarningSuppressOpenCamLib.isChecked())

    def loadSettings(self):
        self.form.WarningSuppressAllSpeeds.setChecked(PathPreferences.suppressAllSpeedsWarning())
        self.form.WarningSuppressRapidSpeeds.setChecked(PathPreferences.suppressRapidSpeedsWarning(False))
        self.form.WarningSuppressSelectionMode.setChecked(PathPreferences.suppressSelectionModeWarning())
        self.form.EnableAdvancedOCLFeatures.setChecked(PathPreferences.advancedOCLFeaturesEnabled())
        self.form.WarningSuppressOpenCamLib.setChecked(PathPreferences.suppressOpenCamLibWarning())
        self.updateSelection()

    def updateSelection(self, state=None):
        self.form.WarningSuppressOpenCamLib.setEnabled(self.form.EnableAdvancedOCLFeatures.isChecked())
        if self.form.WarningSuppressAllSpeeds.isChecked():
            self.form.WarningSuppressRapidSpeeds.setChecked(True)
            self.form.WarningSuppressRapidSpeeds.setEnabled(False)
        else:
            self.form.WarningSuppressRapidSpeeds.setEnabled(True)

